# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# Copyright (C) 2019 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

"""Utilities for DirSrv.

    TODO put them in a module!
"""
import json
import re
import os
import glob
import logging
import shutil
import ldap
import mmap
import socket
import ipaddress
import itertools
import time
import stat
from datetime import (datetime, timedelta)
import sys
import filecmp
import pwd
import shlex
import operator
import subprocess
import math
import errno
from typing import Optional, Union
from socket import getfqdn
from ldapurl import LDAPUrl
from contextlib import closing
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import pkcs12
import lib389
from pathlib import Path
from subprocess import check_output
from lib389.paths import ( Paths, DEFAULTS_PATH )
from lib389._constants import (
        DEFAULT_USER, VALGRIND_WRAPPER, DN_CONFIG, CFGSUFFIX, LOCALHOST,
        ReplicaRole, CONSUMER_REPLICAID, SENSITIVE_ATTRS, DEFAULT_DB_LIB,
        DEFAULT_LMDB_SIZE
    )
from lib389.properties import (
        SER_HOST, SER_USER_ID, SER_GROUP_ID, SER_STRICT_HOSTNAME_CHECKING, SER_PORT,
        SER_ROOT_DN, SER_ROOT_PW, SER_SERVERID_PROP, SER_CREATION_SUFFIX,
        SER_INST_SCRIPTS_ENABLED, SER_SECURE_PORT, REPLICA_ID
    )

MAJOR, MINOR, _, _, _ = sys.version_info

DEBUGGING = os.getenv('DEBUGGING', default=False)

log = logging.getLogger(__name__)

selinux_fcontext_info = None

#
# Various searches to be used in getEntry
#   eg getEntry(*searches['NAMINGCONTEXTS'])
#
searches = {
    'NAMINGCONTEXTS': ('', ldap.SCOPE_BASE, '(objectclass=*)',
                       ['namingcontexts']),
    'ZOMBIE': ('', ldap.SCOPE_SUBTREE,
               '(&(objectclass=glue)(objectclass=extensibleobject))', ['dn'])
}

# Map table for pseudolocalized strings
_chars = {
    " ": u"\u2003",
    "!": u"\u00a1",
    "\"": u"\u2033",
    "#": u"\u266f",
    "$": u"\u20ac",
    "%": u"\u2030",
    "&": u"\u214b",
    "'": u"\u00b4",
    ")": u"}",
    "(": u"{",
    "*": u"\u204e",
    "+": u"\u207a",
    ",": u"\u060c",
    "-": u"\u2010",
    ".": u"\u00b7",
    "/": u"\u2044",
    "0": u"\u24ea",
    "1": u"\u2460",
    "2": u"\u2461",
    "3": u"\u2462",
    "4": u"\u2463",
    "5": u"\u2464",
    "6": u"\u2465",
    "7": u"\u2466",
    "8": u"\u2467",
    "9": u"\u2468",
    ":": u"\u2236",
    ";": u"\u204f",
    "<": u"\u2264",
    "=": u"\u2242",
    ">": u"\u2265",
    "?": u"\u00bf",
    "@": u"\u055e",
    "A": u"\u00c5",
    "B": u"\u0181",
    "C": u"\u00c7",
    "D": u"\u00d0",
    "E": u"\u00c9",
    "F": u"\u0191",
    "G": u"\u011c",
    "H": u"\u0124",
    "I": u"\u00ce",
    "J": u"\u0134",
    "K": u"\u0136",
    "L": u"\u013b",
    "M": u"\u1e40",
    "N": u"\u00d1",
    "O": u"\u00d6",
    "P": u"\u00de",
    "Q": u"\u01ea",
    "R": u"\u0154",
    "S": u"\u0160",
    "T": u"\u0162",
    "U": u"\u00db",
    "V": u"\u1e7c",
    "W": u"\u0174",
    "X": u"\u1e8a",
    "Y": u"\u00dd",
    "Z": u"\u017d",
    "[": u"\u2045",
    "\\": u"\u2216",
    "]": u"\u2046",
    "^": u"\u02c4",
    "_": u"\u203f",
    "`": u"\u2035",
    "a": u"\u00e5",
    "b": u"\u0180",
    "c": u"\u00e7",
    "d": u"\u00f0",
    "e": u"\u00e9",
    "f": u"\u0192",
    "g": u"\u011d",
    "h": u"\u0125",
    "i": u"\u00ee",
    "j": u"\u0135",
    "k": u"\u0137",
    "l": u"\u013c",
    "m": u"\u0271",
    "n": u"\u00f1",
    "o": u"\u00f6",
    "p": u"\u00fe",
    "q": u"\u01eb",
    "r": u"\u0155",
    "s": u"\u0161",
    "t": u"\u0163",
    "u": u"\u00fb",
    "v": u"\u1e7d",
    "w": u"\u0175",
    "x": u"\u1e8b",
    "y": u"\u00fd",
    "z": u"\u017e",
    "{": u"(",
    "}": u")",
    "|": u"\u00a6",
    "~": u"\u02de",
}

#
# Size parser constants
#
SIZE_UNITS = { 't': 2**40, 'g': 2**30, 'm': 2**20, 'k': 2**10, '': 1, }
SIZE_PATTERN = r'\s*(\d*\.?\d*)\s*([tgmk]?)b?\s*'

RPM_TOOL = '/usr/bin/rpm'
LDD_TOOL = '/usr/bin/ldd'
OBJDUMP_TOOL = '/usr/bin/objdump'

#
# Utilities
#


def selinux_present():
    """
    Determine if SELinux libraries are on a system, and if so, if we are in
    a state to consume them (enabled, disabled).

    :returns: bool
    """
    status = False

    try:
        import selinux
        if selinux.is_selinux_enabled():
            # We have SELinux, lets see if we are allowed to configure it.
            # (just checking the uid for now - we may rather check if semanage command success)
            if os.geteuid() != 0:
                log.info('Non-privileged user cannot use semanage, will not relabel ports or files.' )
            elif not os.path.exists('/usr/sbin/semanage'):
                log.info('semanage command is not installed, will not relabel ports or files. Please install policycoreutils-python-utils.' )
            else:
                status = True
        else:
            # We have the module, but it's disabled.
            log.info('SELinux is disabled, will not relabel ports or files.' )
    except ImportError:
        # No python module, so who knows what state we are in.
        log.error('SELinux python module not found, will not relabel files.' )

    return status


def selinux_restorecon(path):
    """
    Relabel a filesystem rooted at path.

    :param path: The filesystem path to recursively relabel
    :type path: str:
    """

    try:
        import selinux
    except ImportError:
        log.debug('SELinux python module not found, skipping relabel path %s' % path)
        return

    if not selinux.is_selinux_enabled():
        log.debug('SELinux is disabled, skipping relabel path %s' % path)
        return

    try:
        selinux.restorecon(path, recursive=True)
    except:
        log.debug("Failed to run restorecon on: " + path)


def _parse_semanage_fcontexts(cmd, regex=r"^(/[^ ]*)[^:=]+:[^:]*:([^:]*):.*$", reject=None):
    """Parse semanage fcontext -L output
    """
    if reject is None:
        reject = {}
    info = {}
    result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    args = ' '.join(ensure_list_str(result.args))
    stdout = ensure_str(result.stdout)
    stderr = ensure_str(result.stderr)
    if result.returncode:
        log.debug("CMD: {args} returned {result.returncode} STDOUT: {stdout} STDERR: {stderr}")
        return info
    for m in re.finditer(regex, stdout, flags=re.MULTILINE):
        if not m.group(1) in reject:
            info[m.group(1)] = m.group(2)
    return info


def _get_selinux_fcontext_info():
    global selinux_fcontext_info
    if not selinux_fcontext_info:
        local = _parse_semanage_fcontexts(["semanage", "fcontext", "-l", "-C"])
        selinux_fcontext_info = {
            "local"  : local,
            "policy" : _parse_semanage_fcontexts(["semanage", "fcontext", "-l"], reject=local),
            "equiv"  : _parse_semanage_fcontexts(["semanage", "fcontext", "-l"], regex=r"^(/[^ ]*) += +(.*)$"),
        }
    return selinux_fcontext_info


def resolve_selinux_path(path):
    """Return the path as expected by semanage fcontext"""
    path = str(Path(path).resolve())
    if selinux_present():
        _get_selinux_fcontext_info()
        equiv = selinux_fcontext_info['equiv']
        for r in equiv:
            if path.startswith(r):
                path = path.replace(r, equiv[r])
    return path


def selinux_label_file(path, label):
    """
    Set set an SELinux label to a file (or a directory)

    :param path: The file path
    :type path: str
    :param label: The label to set (None to unset a label)
    :type path: str
    :raises: ValueError: Error message
    """
    if not selinux_present():
        return
    _get_selinux_fcontext_info()
    if os.path.exists(path):
        path = resolve_selinux_path(path)
    local = selinux_fcontext_info['local']
    policy = selinux_fcontext_info['policy']
    if path in local:
        if local[path] == label:
            return
        log.debug(f"Removing SELinux file context {path} with label {local[path]}.")
        subprocess.run(["semanage", "fcontext", "-d", path])
        del local[path]
    if path in policy:
        if policy[path] == label:
            return
        raise ValueError(f'Cannot change file context for {path} because it is defined in SELinux policy. Please choose another path.')
    if label:
        rc = 0
        for i in range(5):
            try:
                log.debug(f"Setting label {label} in SELinux file context {path}. Attempt {i}")
                result = subprocess.run(["semanage", "fcontext", "-a", "-t", label, path],
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
                args = ' '.join(ensure_list_str(result.args))
                stdout = ensure_str(result.stdout)
                stderr = ensure_str(result.stderr)
                rc = result.returncode
                if rc == 0:
                    local[path] = label
                    break
                else:
                    log.debug(f"Failure setting label {label} for context {path}: Result code {rc}, retrying...")
                    time.sleep(2)
            except (OSError, subprocess.CalledProcessError) as e:
                log.debug(f"Failure setting label {label} for context {path}: Exception {str(e)}, retrying...")
                time.sleep(2)

        if rc != 0:
            log.error(f"ERROR CMD: {args} ; STDOUT: {stdout} ; STDERR: {stderr}")
            try:
                result.check_returncode()
            except (OSError, subprocess.CalledProcessError) as e:
                raise ValueError(f"Failed to set SElinux label {label} on {path}: {str(e)}")

    if os.path.exists(path):
        # pytest fails if I use selinux_restorecon(path)
        subprocess.run(["restorecon", "-R", path])


def selinux_clean_files_label(all=False):
    """Remove dirsrv_*_t labels from either removed files or all files."""
    if not selinux_present():
        return
    _get_selinux_fcontext_info()
    local = { **selinux_fcontext_info['local'] }
    for path, label in local.items():
        if label in ( 'dirsrv_config_t', 'dirsrv_tmpfs_t', 'dirsrv_var_lib_t', 'dirsrv_var_lock_t', 'dirsrv_var_log_t', 'dirsrv_var_run_t', ):
            if all or not os.path.exists(path):
                selinux_label_file(path, None)


def selinux_clean_ports_label():
    """Remove labels from all port having ldap_port_t labels."""
    if not selinux_present():
        return
    cmd=[ 'semanage', 'port', '--list', '-C' ]
    res = _parse_semanage_fcontexts(cmd, regex=r"^(ldap_port_t) *tcp *(.*)$")
    if 'ldap_port_t' in res:
        for port in res['ldap_port_t'].split(","):
            # Cannot use selinux_label_port here because port may be a range.
            subprocess.run(["semanage", "port", "-d", "-p", "tcp", port.strip()])


def _get_selinux_port_policies(port):
    """Get a list of SELinux port policies for the specified port, 'tcp' protocol and
    excluding 'unreserved_port_t', 'reserved_port_t', 'ephemeral_port_t' labels"""

    # [2:] - cut first lines containing the headers. [:-1] - empty line
    policy_lines = subprocess.check_output(["semanage", "port", "-l"], encoding='utf-8').split("\n")[2:][:-1]
    policies = []
    for line in policy_lines:
        data = line.split()
        ports_list = []
        for p in data[2:]:
            if "," in p:
                p = p[:-1]
            if "-" in p:
                p = range(int(p.split("-")[0]), int(p.split("-")[1]))
            else:
                p = [int(p)]
            ports_list.extend(p)
        if data[1] == 'tcp' and port in ports_list and \
           data[0] not in ['hi_reserved_port_t', 'unreserved_port_t', 'reserved_port_t', 'ephemeral_port_t']:
            policies.append({'protocol': data[1], 'type': data[0], 'ports': ports_list})
    return policies


def selinux_label_port(port, remove_label=False):
    """
    Either set or remove an SELinux label(ldap_port_t) for a TCP port

    :param port: The TCP port to be labeled
    :type port: str
    :param remove_label: Set True if the port label should be removed
    :type remove_label: boolean
    :raises: ValueError: Error message
    """
    if port is None:
        return
    try:
        import selinux
    except ImportError:
        log.debug('SELinux python module not found, skipping port labeling.')
        return

    if not selinux_present():
        log.debug('SELinux is disabled, skipping port relabel')
        return

    # We only label ports that ARE NOT in the default policy that comes with
    # a RH based system.
    port = int(port)
    selinux_default_ports = [389, 636, 3268, 3269, 7389]
    if port in selinux_default_ports:
        log.debug('port {} already in {}, skipping port relabel'.format(port, selinux_default_ports))
        return
    label_set = False
    label_ex = None

    policies = _get_selinux_port_policies(port)

    for policy in policies:
        if "ldap_port_t" == policy['type']:
            label_set = True  # Port already has our label
            if port in policy['ports']:
                # The port is within the range, just return
                return
            break
        elif not remove_label:
            # Port belongs to someone else (bad)
            # This is only an issue during setting a label, not removing a label
            raise ValueError("Port {} was already labeled with: ({})  Please choose a different port number".format(port, policy['type']))

    if (remove_label and label_set) or (not remove_label and not label_set):
        for i in range(5):
            try:
                result = subprocess.run(["semanage", "port",
                                         "-d" if remove_label else "-a",
                                         "-t", "ldap_port_t",
                                         "-p", "tcp", str(port)],
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
                args = ' '.join(ensure_list_str(result.args))
                stdout = ensure_str(result.stdout)
                stderr = ensure_str(result.stderr)
                log.debug(f"CMD: {args} ; STDOUT: {stdout} ; STDERR: {stderr}")
                return
            except (OSError, subprocess.CalledProcessError) as e:
                label_ex = e
                time.sleep(3)
        raise ValueError("Failed to mangle port label: " + str(label_ex))


def is_a_dn(dn, allow_anon=True):
    """Returns True if the given string is a DN, False otherwise."""
    try:
        if dn == "" and allow_anon is True:
            # The empty string is anonymous.
            return True
        if len(ldap.dn.str2dn(dn)) > 0:
            # We have valid components in the dn.
            return True
    except ldap.DECODING_ERROR:
        # An invalid dn was given.
        pass
    except TypeError:
        # An invalid type was passed to be checked
        pass
    return False


def is_dn_parent(parent_dn, child_dn, ):
    """ check if child_dn is a really a child of parent_dn
    :return True - if child directly under parent, otherwise False
    """
    parent_dn = parent_dn.lower()
    child_dn = child_dn.lower()

    # Do some basic validation first
    if not ldap.dn.is_dn(parent_dn) or not ldap.dn.is_dn(child_dn):
        return False
    if parent_dn == child_dn:
        return False

    # Get the DN comps and length
    parent_dn_comps = ldap.dn.str2dn(parent_dn)
    child_dn_comps = ldap.dn.str2dn(child_dn)
    parent_comp_len = len(parent_dn_comps)
    child_comp_len = len(child_dn_comps)

    # Do a little more validation
    if child_comp_len <= parent_comp_len or (child_comp_len - parent_comp_len) != 1:
        return False

    # Okay, finally comparing the base DN Comps to see if they match
    for i, e in reversed(list(enumerate(parent_dn_comps))):
        child_idx = i + (child_comp_len - parent_comp_len)
        if child_dn_comps[child_idx] != e:
            return False

    # If we got here then child is a direct decendent of parent
    return True


def normalizeDN(dn, usespace=False):
    # not great, but will do until we use a newer version of python-ldap
    # that has DN utilities
    try:
        ary = ldap.explode_dn(dn.lower())
    except ldap.DECODING_ERROR:
        raise ValueError(f"Unable to normalize DN '{dn}'")
    joinstr = ","
    if usespace:
        joinstr = ", "
    return joinstr.join(ary)


def escapeDNValue(dn):
    """convert special characters in a DN into LDAPv3 escapes.

     e.g.
    "dc=example,dc=com" -> \\"dc\\=example\\,\\ dc\\=com\\"
    """
    for cc in (' ', '"', '+', ',', ';', '<', '>', '='):
        dn = dn.replace(cc, '\\' + cc)
    return dn


def escapeDNFiltValue(dn):
    '''convert special characters in a DN into LDAPv3 escapes
    for use in search filters'''
    for cc in (' ', '"', '+', ',', ';', '<', '>', '='):
        dn = dn.replace(cc, '\\%x' % ord(cc))
    return dn


def suffixfilt(suffix):
    """Return a filter matching any possible suffix form.

        eg. normalized, escaped, spaced...
    """
    nsuffix = normalizeDN(suffix)
    spacesuffix = normalizeDN(nsuffix, True)
    escapesuffix = escapeDNFiltValue(nsuffix)
    filt = ('(|(cn=%s)(cn=%s)(cn=%s)(cn="%s")(cn="%s")(cn=%s)(cn="%s"))' %
            (escapesuffix, nsuffix, spacesuffix, nsuffix, spacesuffix, suffix,
             suffix))
    return filt


#
# path tools
#
def get_sbin_dir(sroot=None, prefix=None):
    """Return the sbin directory (default /usr/sbin)."""
    if sroot:
        return "%s/bin/slapd/admin/bin" % sroot
    elif prefix and prefix != '/':
        return "%s/sbin" % prefix
    return "/usr/sbin"


def get_bin_dir(sroot=None, prefix=None):
    """Return the sbin directory (default /usr/bin)."""
    if sroot:
        return "%s/bin/slapd/admin/bin" % sroot
    elif prefix and prefix != '/':
        return "%s/bin" % prefix
    return "/usr/bin"


def get_data_dir(prefix=None):
    """Return the shared data directory (default /usr/share/dirsrv/data)."""
    if prefix and prefix != '/':
        return "%s/share/dirsrv/data" % prefix
    return "/usr/share/dirsrv/data"


def get_plugin_dir(prefix=None):
    """
    Return the plugin directory (default /usr/lib64/dirsrv/plugins).
    This should be 64/32bit aware.
    """
    if prefix and prefix != '/':
        # With prefix installations, even 64 bit, it can be under /usr/lib/
        if os.path.exists("%s/usr/lib/dirsrv/plugins" % prefix):
            return "%s/usr/lib/dirsrv/plugins" % prefix
        else:
            if os.path.exists("%s/usr/lib64/dirsrv/plugins" % prefix):
                return "%s/usr/lib64/dirsrv/plugins" % prefix

    # Need to check for 32/64 bit installations
    if not os.path.exists("/usr/lib64/dirsrv/plugins"):
        if os.path.exists("/usr/lib/dirsrv/plugins"):
            return "/usr/lib/dirsrv/plugins"
    return "/usr/lib64/dirsrv/plugins"


#
# valgrind functions
#
def valgrind_enable(sbin_dir, wrapper=None):
    '''
    Copy the valgrind ns-slapd wrapper into the /sbin directory
    (making a backup of the original ns-slapd binary).

    The script calling valgrind_enable() must be run as the 'root' user
    as SELinux needs to be disabled for valgrind to work

    The server instance(s) should be stopped prior to calling this function.
    Then after calling valgrind_enable():
    - Start the server instance(s) with a timeout of 60 (valgrind takes a
      while to startup)
    - Run the tests
    - Stop the server
    - Get the results file
    - Run valgrind_check_file(result_file, "pattern", "pattern", ...)
    - Run valgrind_disable()

    :param sbin_dir: the location of the ns-slapd binary (e.g. /usr/sbin)
    :param wrapper: The valgrind wrapper script for ns-slapd (if not set,
                     a default wrapper is used)
    :raise IOError: If there is a problem setting up the valgrind scripts
    :raise EnvironmentError: If script is not run as 'root'
    '''

    if not os.access(sbin_dir, os.W_OK):
        # Note: valgrind has no limitation but  ns-slapd must be replaced
        # This check allows non root user to use custom install prefix
        log.error('This script must be run as root to use valgrind (Should at least be able to write in {sbin_dir})')
        raise EnvironmentError

    if not wrapper:
        # use the default ns-slapd wrapper
        wrapper = '%s/%s' % (os.path.dirname(os.path.abspath(__file__)),
                             VALGRIND_WRAPPER)

    nsslapd_orig = '%s/ns-slapd' % sbin_dir
    nsslapd_backup = '%s/ns-slapd.original' % sbin_dir

    if os.path.isfile(nsslapd_backup):
        # There is a backup which means we never cleaned up from a previous
        # run(failed test?)
        if not filecmp.cmp(nsslapd_backup, nsslapd_orig):
            # Files are different sizes, we assume valgrind is already setup
            log.info('Valgrind is already enabled.')
            return

    # Check both nsslapd's exist
    if not os.path.isfile(wrapper):
        raise IOError('The valgrind wrapper (%s) does not exist. file=%s' %
                      (wrapper, __file__))

    if not os.path.isfile(nsslapd_orig):
        raise IOError('The binary (%s) does not exist or is not accessible.' %
                      nsslapd_orig)

    # Make a backup of the original ns-slapd and copy the wrapper into place
    try:
        shutil.copy2(nsslapd_orig, nsslapd_backup)
    except IOError as e:
        log.fatal('valgrind_enable(): failed to backup ns-slapd, error: %s' %
                  e.strerror)
        raise IOError('failed to backup ns-slapd, error: %s' % e.strerror)

    # Copy the valgrind wrapper into place
    try:
        shutil.copy2(wrapper, nsslapd_orig)
    except IOError as e:
        log.fatal('valgrind_enable(): failed to copy valgrind wrapper '
                  'to ns-slapd, error: %s' % e.strerror)
        raise IOError('failed to copy valgrind wrapper to ns-slapd, error: %s' %
                      e.strerror)

    # Disable SELinux
    if os.geteuid() == 0:
        os.system('setenforce 0')

    # Disable systemd by turning off with_system in .inf file
    old_path = Paths()._get_defaults_loc(DEFAULTS_PATH)
    new_path = f'{old_path}.orig'
    os.rename(old_path, new_path)
    with open(new_path, 'rt') as fin:
       with open(old_path, 'wt') as fout:
            for line in fin:
                if line.startswith('with_systemd'):
                    fout.write('with_systemd = 0\n')
                else:
                    fout.write(line)

    log.info('Valgrind is now enabled.')


def valgrind_disable(sbin_dir):
    '''
    Restore the ns-slapd binary to its original state - the server instances
    are expected to be stopped.

    Note - SELinux is enabled at the end of this process.

    :param sbin_dir - the location of the ns-slapd binary (e.g. /usr/sbin)
    :raise ValueError
    :raise EnvironmentError: If script is not run as 'root'
    '''

    if not os.access(sbin_dir, os.W_OK):
        # Note: valgrind has no limitation but  ns-slapd must be replaced
        # This check allows non root user to use custom install prefix
        log.error('This script must be run as root to use valgrind (Should at least be able to write in {sbin_dir})')
        raise EnvironmentError

    nsslapd_orig = '%s/ns-slapd' % sbin_dir
    nsslapd_backup = '%s/ns-slapd.original' % sbin_dir

    # Restore the original ns-slapd
    try:
        shutil.copyfile(nsslapd_backup, nsslapd_orig)
    except IOError as e:
        log.fatal('valgrind_disable: failed to restore ns-slapd, error: %s' %
                  e.strerror)
        raise ValueError('failed to restore ns-slapd, error: %s' % e.strerror)

    # Delete the backup now
    try:
        os.remove(nsslapd_backup)
    except OSError as e:
        log.fatal('valgrind_disable: failed to delete backup ns-slapd, error:'
                  ' %s' % e.strerror)
        raise ValueError('Failed to delete backup ns-slapd, error: %s' %
                         e.strerror)

    # Enable SELinux
    if os.geteuid() == 0:
        os.system('setenforce 1')

    # Restore .inf file (for systemd)
    new_path = Paths()._get_defaults_loc(DEFAULTS_PATH)
    old_path = f'{new_path}.orig'
    if os.path.exists(old_path):
        os.replace(old_path, new_path)

    log.info('Valgrind is now disabled.')


def valgrind_get_results_file(dirsrv_inst):
    """
    Return the valgrind results file for the dirsrv instance.
    """

    """
    The "ps -ef | grep valgrind" looks like:

        nobody 26239 1 10 14:33 ? 00:00:06 valgrind -q --tool=memcheck
        --leak-check=yes --leak-resolution=high --num-callers=50
        --log-file=/var/tmp/slapd.vg.26179 /usr/sbin/ns-slapd.original
        -D /etc/dirsrv/slapd-localhost -i /var/run/dirsrv/slapd-localhost.pid
        -w /var/run/dirsrv/slapd-localhost.startpid

    We need to extract the "--log-file" value
    """
    cmd = ("ps -ef | grep valgrind | grep 'slapd-" + dirsrv_inst.serverid +
           " ' | awk '{ print $14 }' | sed -e 's/\\-\\-log\\-file=//'")

    # Run the command and grab the output
    p = os.popen(cmd)
    results_file = p.readline().strip()
    p.close()

    return results_file


def valgrind_check_file(results_file, *patterns):
    '''
    Check the valgrind results file for the all the patterns
    @param result_file - valgrind results file (must be read after server is
                         stopped)
    @param patterns - A plain text or regex pattern string args that should
                      be searched for
    @return True/False - Return true if one if the patterns match a stack
                         trace
    @raise IOError
    '''

    # Verify results file
    if not results_file:
        assert False

    # Check the result file fo the leak text
    results_file = results_file.replace('\n', '')
    found = False
    pattern_count = len(patterns)
    matched_count = 0

    vlog = open(results_file)
    for line in vlog:
        for match_txt in patterns:
            if re.search(match_txt, line):
                matched_count += 1
                break

        if len(line.split()) == 1:
            # Check if this stack stack matched all the patterns
            if pattern_count == matched_count:
                found = True
                print('valgrind: match found in results file: %s' %
                      (results_file))
                break
            else:
                matched_count = 0
    vlog.close()

    return found


#
# functions using sockets
#
def isLocalHost(host_name):
    """True if host_name points to a local ip.

        Uses gethostbyname()
    """
    # first see if this is a "well known" local hostname
    if host_name is None or \
       host_name == 'localhost' or \
       host_name == 'localhost.localdomain' or \
       host_name == socket.gethostname():
        return True

    # first lookup ip addr
    try:
        ip_addr = socket.gethostbyname(host_name)
        if ip_addr.startswith("127."):
            # log.debug("this ip is on loopback, retain only the first octet")
            ip_addr = '127.'
    except socket.gaierror:
        log.debug("no ip address for %r" % host_name)
        return False

    # next, see if this IP addr is one of our
    # local addresses
    p = subprocess.Popen(['/sbin/ip', 'addr'], stdout=subprocess.PIPE)
    child_stdout = p.stdout.read()
    found = ('inet %s' % ip_addr).encode() in child_stdout
    p.wait()

    return found


def getdomainname(name=''):
    fqdn = getfqdn(name)
    index = fqdn.find('.')
    if index >= 0:
        return fqdn[index + 1:]
    else:
        # This isn't correct. There is no domain name, so return empty.str
        # return fqdn
        return ""


def getdefaultsuffix(name=''):
    dm = getdomainname(name)
    if dm:
        return "dc=" + dm.replace('.', ',dc=')
    else:
        return 'dc=localdomain'


def get_server_user(args):
    """Return the unix username used from the server inspecting the following
       keys in args.

        'newuserid', 'admconf', 'sroot' -> ssusers.conf

    """
    if 'newuserid' not in args:
        if 'admconf' in args:
            args['newuserid'] = args['admconf'].SuiteSpotUserID
        elif 'sroot' in args:
            ssusers = open("%s/shared/config/ssusers.conf" % args['sroot'])
            for line in ssusers:
                ary = line.split()
                if len(ary) > 1 and ary[0] == 'SuiteSpotUser':
                    args['newuserid'] = ary[-1]
            ssusers.close()
    if 'newuserid' not in args:
        args['newuserid'] = os.environ['LOGNAME']
        if args['newuserid'] == 'root':
            args['newuserid'] = DEFAULT_USER


def update_newhost_with_fqdn(args):
    """Replace args[SER_HOST] with its fqdn and returns True if local.

    One of the arguments to createInstance is newhost.  If this is specified,
    we need to convert it to the fqdn.  If not given, we need to figure out
    what the fqdn of the local host is.  This method sets newhost in args to
    the appropriate value and returns True if newhost is the localhost, False
    otherwise"""
    if SER_HOST in args:
        args[SER_HOST] = getfqdn(args[SER_HOST])
        isLocal = isLocalHost(args[SER_HOST])
    else:
        isLocal = True
        args[SER_HOST] = getfqdn()
    return isLocal


def getcfgdsuserdn(cfgdn, args):
    """Return a DirSrv object bound anonymously or to the admin user.

    If the config ds user ID was given, not the full DN, we need to figure
    out the full DN.

    Try in order to:
        1- search the directory anonymously;
        2- look in ldap.conf;
        3- try the default DN.

    This may raise a file or LDAP exception.
    """
    # create a connection to the cfg ds
    conn = lib389.DirSrv(args['cfgdshost'], args['cfgdsport'], "", "", None)
    # if the caller gave a password, but not the cfguser DN, look it up
    if 'cfgdspwd' in args and \
            ('cfgdsuser' not in args or not is_a_dn(args['cfgdsuser'])):
        if 'cfgdsuser' in args:
            ent = conn.getEntry(cfgdn, ldap.SCOPE_SUBTREE,
                                "(uid=%s)" % args['cfgdsuser'],
                                ['dn'])
            args['cfgdsuser'] = ent.dn
        elif 'sroot' in args:
            ldapconf = open(
                "%s/shared/config/ldap.conf" % args['sroot'], 'r')
            for line in ldapconf:
                ary = line.split()  # default split is all whitespace
                if len(ary) > 1 and ary[0] == 'admnm':
                    args['cfgdsuser'] = ary[-1]
            ldapconf.close()
        elif 'admconf' in args:
            args['cfgdsuser'] = args['admconf'].userdn
        elif 'cfgdsuser' in args:
            dsuser = ("uid=%s,ou=Administrators,ou=TopologyManagement,%s" %
                      (args['cfgdsuser'], cfgdn))
            args['cfgdsuser'] = dsuser
        conn.unbind()
        conn = lib389.DirSrv(
            args['cfgdshost'], args['cfgdsport'], args['cfgdsuser'],
            args['cfgdspwd'], None)
    return conn


def update_admin_domain(isLocal, args):
    """Get the admin domain to use."""
    if isLocal and 'admin_domain' not in args:
        if 'admconf' in args:
            args['admin_domain'] = args['admconf'].admindomain
        elif 'sroot' in args:
            dsconf = open('%s/shared/config/ds.conf' % args['sroot'], 'r')
            for line in dsconf:
                ary = line.split(":")
                if len(ary) > 1 and ary[0] == 'AdminDomain':
                    args['admin_domain'] = ary[1].strip()
            dsconf.close()


def getoldcfgdsinfo(args):
    """Use the old style sroot/shared/config/dbswitch.conf to get the info"""
    dbswitch = open("%s/shared/config/dbswitch.conf" % args['sroot'], 'r')
    try:
        matcher = re.compile(r'^directory\s+default\s+')
        for line in dbswitch:
            m = matcher.match(line)
            if m:
                url = LDAPUrl(line[m.end():])
                ary = url.hostport.split(":")
                if len(ary) < 2:
                    ary.append(389)
                else:
                    ary[1] = int(ary[1])
                ary.append(url.dn)
                return ary
    finally:
        dbswitch.close()


def getnewcfgdsinfo(new_instance_arguments):
    """Use the new style prefix /etc/dirsrv/admin-serv/adm.conf.

        new_instance_arguments = {'admconf': obj } where obj.ldapurl != None
    """
    try:
        url = LDAPUrl(new_instance_arguments['admconf'].ldapurl)
    except AttributeError:
        log.error("missing ldapurl attribute in new_instance_arguments: %r" %
                  new_instance_arguments)
        raise

    ary = url.hostport.split(":")
    if len(ary) < 2:
        ary.append(389)
    else:
        ary[1] = int(ary[1])
    ary.append(url.dn)
    return ary


def getcfgdsinfo(new_instance_arguments):
    """Returns a 3-tuple consisting of the host, port, and cfg suffix.

        `new_instance_arguments` = {
            'cfgdshost':
            'cfgdsport':
            'new_style':
        }
    We need the host and port of the configuration directory server in order
    to create an instance.  If this was not given, read the dbswitch.conf file
    to get the information.  This method will raise an exception if the file
    was not found or could not be open.  This assumes new_instance_arguments
    contains the sroot parameter for the server root path.
    """
    try:
        return (new_instance_arguments['cfgdshost'],
                int(new_instance_arguments['cfgdsport']),
                CFGSUFFIX)
    except KeyError:  # if keys are missing...
        if new_instance_arguments['new_style']:
            return getnewcfgdsinfo(new_instance_arguments)

        return getoldcfgdsinfo(new_instance_arguments)


def getserverroot(cfgconn, isLocal, args):
    """Grab the serverroot from the instance dir of the config ds if the user
    did not specify a server root directory"""
    if cfgconn and 'sroot' not in args and isLocal:
        ent = cfgconn.getEntry(
            DN_CONFIG, ldap.SCOPE_BASE, "(objectclass=*)",
            ['nsslapd-instancedir'])
        if ent:
            args['sroot'] = os.path.dirname(
                ent.getValue('nsslapd-instancedir'))


def getadminport(cfgconn, cfgdn, args):
    """Return a 2-tuple (asport, True) if the admin server is using SSL,
    False otherwise.

    Get the admin server port so we can contact it via http.  We get this from
    the configuration entry using the CFGSUFFIX and cfgconn.  Also get any
    other information we may need from that entry.
    """
    asport = 0
    secure = False
    if cfgconn:
        dn = cfgdn
        if 'admin_domain' in args:
            dn = "cn=%s,ou=%s, %s" % (
                args[SER_HOST], args['admin_domain'], cfgdn)
        filt = "(&(objectclass=nsAdminServer)(serverHostName=%s)" % args[
            SER_HOST]
        if 'sroot' in args:
            filt += "(serverRoot=%s)" % args['sroot']
        filt += ")"
        ent = cfgconn.getEntry(
            dn, ldap.SCOPE_SUBTREE, filt, ['serverRoot'])
        if ent:
            if 'sroot' not in args and ent.serverRoot:
                args['sroot'] = ent.serverRoot
            if 'admin_domain' not in args:
                ary = ldap.explode_dn(ent.dn, 1)
                args['admin_domain'] = ary[-2]
            dn = "cn=configuration, " + ent.dn
            ent = cfgconn.getEntry(dn, ldap.SCOPE_BASE, '(objectclass=*)',
                                   ['nsServerPort',
                                    'nsSuiteSpotUser',
                                    'nsServerSecurity'])
            if ent:
                asport = ent.nsServerPort
                secure = (ent.nsServerSecurity and (
                    ent.nsServerSecurity == 'on'))
                if 'newuserid' not in args:
                    args['newuserid'] = ent.nsSuiteSpotUser
        cfgconn.unbind()
    return asport, secure


def formatInfData(args):
    """Return the .inf data for a silence setup via setup-ds.pl.

        args = {
            # new instance values
            newhost, newuserid, newport, SER_ROOT_DN, newrootpw, newsuffix,

            # The following parameters require to register the new instance
            # in the admin server
            have_admin, cfgdshost, cfgdsport, cfgdsuser,cfgdspwd, admin_domain

            InstallLdifFile, AddOrgEntries, ConfigFile, SchemaFile,
            ldapifilepath

            # Setup the o=NetscapeRoot namingContext
            setup_admin,
        }

        @see https://access.redhat.com/site/documentation/en-US/
             Red_Hat_Directory_Server/8.2/html/Installation_Guide/
             Installation_Guide-Advanced_Configuration-Silent.html
        [General]
        FullMachineName= dir.example.com
        SuiteSpotUserID= nobody
        SuiteSpotGroup= nobody
        AdminDomain= example.com
        ConfigDirectoryAdminID= admin
        ConfigDirectoryAdminPwd= admin
        ConfigDirectoryLdapURL= ldap://dir.example.com:389/o=NetscapeRoot

        [slapd]
        SlapdConfigForMC= Yes
        UseExistingMC= 0
        ServerPort= 389
        ServerIdentifier= dir
        Suffix= dc=example,dc=com
        RootDN= cn=Directory Manager
        RootDNPwd= password
        ds_bename=exampleDB
        AddSampleEntries= No

        [admin]
        Port= 9830
        ServerIpAddress= 111.11.11.11
        ServerAdminID= admin
        ServerAdminPwd= admin

    """
    args = args.copy()
    args['CFGSUFFIX'] = CFGSUFFIX

    content = ("[General]" "\n")
    content += ("FullMachineName= %s\n" % args[SER_HOST])
    content += ("SuiteSpotUserID= %s\n" % args[SER_USER_ID])
    content += ("SuiteSpotGroup= %s\n" % args[SER_GROUP_ID])
    content += ("StrictHostCheck= %s\n" % args[SER_STRICT_HOSTNAME_CHECKING])

    if args.get('have_admin'):
        content += ("AdminDomain= %(admin_domain)s" "\n"
                    "ConfigDirectoryLdapURL= ldap://%(cfgdshost)s:%"
                    "(cfgdsport)d/%(CFGSUFFIX)s" "\n"
                    "ConfigDirectoryAdminID= %(cfgdsuser)s" "\n"
                    "ConfigDirectoryAdminPwd= %(cfgdspwd)s" "\n") % args

    content += ("\n" "\n" "[slapd]" "\n")
    content += ("ServerPort= %s\n" % args[SER_PORT])
    content += ("RootDN= %s\n" % args[SER_ROOT_DN])
    content += ("RootDNPwd= %s\n" % args[SER_ROOT_PW])
    content += ("ServerIdentifier= %s\n" % args[SER_SERVERID_PROP])
    content += ("Suffix= %s\n" % args[SER_CREATION_SUFFIX])

    # Create admin?
    if args.get('setup_admin'):
        content += ("SlapdConfigForMC= Yes" "\n"
                    "UseExistingMC= 0 " "\n")

    if 'InstallLdifFile' in args:
        content += """\nInstallLdifFile= %s\n""" % args['InstallLdifFile']
    if 'AddOrgEntries' in args:
        content += """\nAddOrgEntries= %s\n""" % args['AddOrgEntries']
    if 'ConfigFile' in args:
        for ff in args['ConfigFile']:
            content += """\nConfigFile= %s\n""" % ff
    if 'SchemaFile' in args:
        for ff in args['SchemaFile']:
            content += """\nSchemaFile= %s\n""" % ff

    if 'ldapifilepath' in args:
        content += "\nldapifilepath=%s\n" % args['ldapifilepath']
    if SER_INST_SCRIPTS_ENABLED in args:
        content += "\n{}={}\n".format(SER_INST_SCRIPTS_ENABLED, args[SER_INST_SCRIPTS_ENABLED])

    return content


def generate_ds_params(inst_num, role=ReplicaRole.STANDALONE):
    """Generate a host, port, secure port, server ID and replica ID
    for the selected role and instance number. I.e. inst_num=1, role="supplier".

    @param inst_num - an instance number in a range from 1 to 99
    @param role - ReplicaRole.STANDALONE, ReplicaRole.SUPPLIER, ReplicaRole.HUB, ReplicaRole.CONSUMER
    @return - the dict with next keys: host, port, secure port, server id and replica id
    """

    if inst_num not in range(1, 100):
        raise ValueError("Instance number should be in a range from 1 to 99")

    if role not in (ReplicaRole.STANDALONE, ReplicaRole.SUPPLIER, ReplicaRole.HUB, ReplicaRole.CONSUMER):
        raise ValueError('Role should be {}, {}, {}, {}'.format(ReplicaRole.STANDALONE, ReplicaRole.SUPPLIER,
                                                                ReplicaRole.HUB, ReplicaRole.CONSUMER))

    instance_data = {}
    relevant_num = 38900

    # Set relevant number for ports
    if role == ReplicaRole.SUPPLIER:
        relevant_num += 100
    elif role == ReplicaRole.HUB:
        relevant_num += 200
    elif role == ReplicaRole.CONSUMER:
        relevant_num += 300

    # Define replica ID
    if role == ReplicaRole.SUPPLIER:
        replica_id = inst_num
    else:
        replica_id = CONSUMER_REPLICAID

    # Fill the dict with data
    server_id = "{}{}".format(role.name.lower(), inst_num)
    instance_data[SER_HOST] = LOCALHOST
    instance_data[SER_PORT] = relevant_num + inst_num
    instance_data[SER_SECURE_PORT] = relevant_num + inst_num + 24700
    instance_data[SER_SERVERID_PROP] = server_id
    instance_data[REPLICA_ID] = replica_id

    return instance_data

class DSVersion():
    def __init__(self, version):
        self._version = str(version)
        self._key = _cmpkey(self._version)

    def __str__(self):
        return self._version

    def __repr__(self):
        return f"<DSVersion('{self}')>"

    def __hash__(self):
        return hash(self._key)

    def __lt__(self, other):
        if not isinstance(other, DSVersion):
            return NotImplemented

        return self._key < other._key

    def __le__(self, other):
        if not isinstance(other, DSVersion):
            return NotImplemented

        return self._key <= other._key

    def __eq__(self, other):
        if not isinstance(other, DSVersion):
            return NotImplemented

        return self._key == other._key

    def __ge__(self, other):
        if not isinstance(other, DSVersion):
            return NotImplemented

        return self._key >= other._key

    def __gt__(self, other):
        if not isinstance(other, DSVersion):
            return NotImplemented

        return self._key > other._key

    def __ne__(self, other):
        if not isinstance(other, DSVersion):
            return NotImplemented

        return self._key != other._key


def _parse_version_parts(s):
    for part in re.compile(r"(\d+ | [a-z]+ | \. | -)", re.VERBOSE).split(s):

        if not part or part == ".":
            continue

        if part[:1] in "0123456789":
            # pad for numeric comparison
            yield part.zfill(8)
        else:
            yield "*" + part

def _cmpkey(version):
    parts = []
    for part in _parse_version_parts(version.lower()):
        parts.append(part)

    return tuple(parts)


def get_ds_version(paths=None):
    """
    Return version of ns-slapd installed on this system. This is determined by the defaults.inf
    file, so is correct for the built and installed ns-slapd binary. This function works without
    root permissions.

    returns a string of the form: 1.3.4.8
    """
    if paths is None:
        paths = Paths()
    return paths.version


def ds_is_related(relation, *ver, instance=None):
    """
    Return a result of a comparison between the current version of ns-slapd and a provided version.
    """
    ops = {'older': operator.lt,
           'newer': operator.ge}
    if instance is None:
        ds_ver = get_ds_version()
    else:
        ds_ver = get_ds_version(instance.ds_paths)
    if len(ver) > 1:
        for cmp_ver in ver:
            if cmp_ver.startswith(ds_ver[:3]):
                return ops[relation](DSVersion(ds_ver), DSVersion(cmp_ver))
    else:
        return ops[relation](DSVersion(ds_ver), DSVersion(ver[0]))


def ds_is_older(*ver, instance=None):
    """
    Return True if the current version of ns-slapd is older than a provided version
    """
    return ds_is_related('older', *ver, instance=instance)


def ds_is_newer(*ver, instance=None):
    """
    Return True if the current version of ns-slapd is newer than a provided version
    """
    return ds_is_related('newer', *ver, instance=instance)

def ds_supports_new_changelog():
    """
    Return True if the current version of ns-slapd supports changelogs under cn=changelog,cn=<backend>,cn=ldbm..
    """
    return ds_is_newer('1.4.4.3')

def gentime_to_datetime(gentime):
    """Convert Generalized time to datetime object

    :param gentime: Time in the format - YYYYMMDDHHMMSSZ (20170126120000Z)
    :type password: str
    :returns: datetime.datetime object
    """

    return datetime.strptime(gentime, '%Y%m%d%H%M%SZ')


def gentime_to_posix_time(gentime):
    """Convert Generalized time to POSIX time format

    :param gentime: Time in the format - YYYYMMDDHHMMSSZ (20170126120000Z)
    :type password: str
    :returns: Epoch time int
    """
    time_tuple = (int(gentime[:4]),
                  int(gentime[4:6]),
                  int(gentime[6:8]),
                  int(gentime[8:10]),
                  int(gentime[10:12]),
                  int(gentime[12:14]),
                  0, 0, 0)

    return time.mktime(time_tuple)


def getDateTime():
    """
    Return the date and time:

        2016-04-21 21:21:00

    """
    return time.strftime("%Y-%m-%d %H:%M:%S")


def socket_check_open(host, port):
    """
    Check if a socket can be opened.  Need to handle cases where IPv6 is completely
    disabled.
    """
    try:
        # Trying IPv6 first ...
        family = socket.AF_INET6
        with closing(socket.socket(family, socket.SOCK_STREAM)) as sock:
            if sock.connect_ex((host, port)) == 0:
                return True
            else:
                return False
    except OSError:
        # No IPv6, adjust hostname, and try IPv4 ...
        family = socket.AF_INET
        if host == "::1":
            host = "127.0.0.1"
        with closing(socket.socket(family, socket.SOCK_STREAM)) as sock:
            if sock.connect_ex((host, port)) == 0:
                return True
            else:
                return False


def socket_check_bind(port):
    """
    Check if a socket can be bound.  Need to handle cases where IPv6 is
    completely disabled.
    """
    # Trying IPv6 first ...
    host = "::1"
    family = socket.AF_INET6
    try:
        with closing(socket.socket(family, socket.SOCK_STREAM)) as sock:
            sock.bind((host, port))
            return True
    except OSError as e:
        if e.errno == errno.EACCES or  e.errno == errno.EADDRINUSE:
            return False
    # Maybe there is no IPv6, adjust hostname, and try IPv4 ...
    host = "127.0.0.1"
    family = socket.AF_INET
    try:
        with closing(socket.socket(family, socket.SOCK_STREAM)) as sock:
            sock.bind((host, port))
            return True
    except OSError as e:
        return False


def ensure_bytes(val):
    if val is not None and type(val) != bytes:
        return val.encode()
    return val


def ensure_str(val):
    if val is not None and type(val) != str:
        try:
            result = val.decode('utf-8')
        except UnicodeDecodeError:
            # binary value, just return str repr?
            result = str(val)
        return result
    return val


def ensure_int(val):
    if val is not None and not isinstance(val, int):
        return int(val)
    return val


def ensure_list_bytes(val):
    return [ensure_bytes(v) for v in val]


def ensure_list_str(val):
    return [ensure_str(v) for v in val]


def ensure_list_int(val):
    return [ensure_int(v) for v in val]


def ensure_dict_str(val):
    if MAJOR <= 2:
        return val
    retdict = {}
    for k in val:
        if isinstance(val[k], list):
            retdict[k] = ensure_list_str(val[k])
        else:
            retdict[k] = ensure_str(val[k])
    return retdict


def align_to_page_size(size):
    page_size = os.sysconf("SC_PAGE_SIZE")
    return ((size + page_size - 1) // page_size) * page_size


def pseudolocalize(string):
    pseudo_string = u""
    for char in string:
        try:
            pseudo_string += _chars[char]
        except KeyError:
            pseudo_string += char
    return pseudo_string


def assert_c(condition, msg="Assertion Failed"):
    """This is the same as assert, but assert is compiled out
    when optimisation is enabled. This prevents compiling out.
    """
    if not condition:
        raise AssertionError(msg)


def format_cmd_list(cmd):
    """Format the subprocess command list to the quoted shell string"""

    return " ".join(map(shlex.quote, cmd))


def get_instance_list():
    # List all server instances
    paths = Paths()
    conf_dir = os.path.join(paths.sysconf_dir, 'dirsrv')
    insts = []
    try:
        for inst in os.listdir(conf_dir):
            if inst.startswith('slapd-') and not inst.endswith('.removed'):
                insts.append(inst)
    except OSError as e:
        log.error("Failed to check directory: {} - {}".format(conf_dir, str(e)))

    insts.sort()
    return insts


# used by arg parse completers
def instance_choices(prefix, parsed_args, **kwargs):
    insts = get_instance_list()
    inst_names = []
    for inst in insts:
        inst_names.append(inst)
        inst_names.append(inst.replace("slapd-", ""))

    return inst_names


def get_user_is_ds_owner():
    # Check if we have permission to administer the DS instance. This is required
    # for some tasks such as installing, killing, or editing configs for the
    # instance.
    cur_uid = os.getuid()
    if cur_uid == 0:
        # We are root, we have permission
        return True
    cur_username = pwd.getpwuid(cur_uid)[0]
    p = Paths()
    if cur_username == p.user:
        # We are the same user, all good
        return True
    return False


def get_user_is_root():
    cur_uid = os.getuid()
    if cur_uid == 0:
        # We are root, we have permission
        return True
    return False


def basedn_to_ldap_dns_uri(basedn):
    # ldap:///dc%3Dexample%2Cdc%3Dcom
    return "ldaps:///" + basedn.replace("=", "%3D").replace(",", "%2C")


def display_log_value(attr, value, hide_sensitive=True):
    # Mask all the sensitive attribute values
    if DEBUGGING or not hide_sensitive:
        return value
    else:
        if attr.lower() in SENSITIVE_ATTRS:
            if type(value) in (list, tuple):
                return list(map(lambda _: '********', value))
            else:
                return '********'
        else:
            return value


def display_log_data(data, hide_sensitive=True):
    # Take a dict and mask all the sensitive data
    return {a: display_log_value(a, v, hide_sensitive) for a, v in data.items()}


def convert_bytes(bytes):
    bytes = int(bytes)
    if bytes == 0:
        return "0 B"
    size_name = ["B", "KB", "MB", "GB", "TB", "PB"]
    i = int(math.floor(math.log(bytes, 1024)))
    pow = math.pow(1024, i)
    siz = round(bytes / pow, 2)
    return "{} {}".format(siz, size_name[i])


def search_filter_escape_bytes(bytes_value):
    """ Convert a byte sequence to a properly escaped for LDAP (format BACKSLASH HEX HEX) string"""
    # copied from https://github.com/cannatag/ldap3/blob/master/ldap3/utils/conv.py
    if str is not bytes:
        if isinstance(bytes_value, str):
            bytes_value = bytearray(bytes_value, encoding='utf-8')
        return ''.join([('\\%02x' % int(b)) for b in bytes_value])
    else:
        raise RuntimeError('Running with Python 2 is unsupported')


def print_nice_time(seconds):
    """Convert seconds to a pretty format
    """
    seconds = int(seconds)
    d, s = divmod(seconds, 24*60*60)
    h, s = divmod(s, 60*60)
    m, s = divmod(s, 60)
    d_plural = ""
    h_plural = ""
    m_plural = ""
    s_plural = ""
    if d > 1:
        d_plural = "s"
    if h != 1:
        h_plural = "s"
    if m != 1:
        m_plural = "s"
    if s != 1:
        s_plural = "s"
    if d > 0:
        return f'{d:d} day{d_plural}, {h:d} hour{h_plural}, {m:d} minute{m_plural}, {s:d} second{s_plural}'
    elif h > 0:
        return f'{h:d} hour{h_plural}, {m:d} minute{m_plural}, {s:d} second{s_plural}'
    elif m > 0:
        return f'{m:d} minute{m_plural}, {s:d} second{s_plural}'
    else:
        return f'{s:d} second{s_plural}'


def copy_with_permissions(source, target):
    """Copy a file while preserving all permissions"""

    shutil.copy2(source, target)
    st = os.stat(source)
    os.chown(target, st[stat.ST_UID], st[stat.ST_GID])
    os.chmod(target, st[stat.ST_MODE])


def cmp(self, x, y):
    """
    Replacement for built-in function cmp that was removed in Python 3

    Compare the two objects x and y and return an integer according to
    the outcome. The return value is negative if x < y, zero if x == y
    and strictly positive if x > y.
    """
    return (x > y) - (x < y)


def is_valid_hostname(hostname):
    if len(hostname) > 255:
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1] # strip exactly one dot from the right, if present
    allowed = re.compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in hostname.split("."))

def is_valid_ip(ip):
    """ Validate an IPv4 or IPv6 address, including asterisks for wildcards. """
    if '*' in ip and '.' in ip:
        ipv4_pattern = r'^(\d{1,3}|\*)\.(\d{1,3}|\*)\.(\d{1,3}|\*)\.(\d{1,3}|\*)$'
        if re.match(ipv4_pattern, ip):
            octets = ip.split('.')
            for octet in octets:
                if octet != '*':
                    try:
                        val = int(octet, 10)
                        if not (0 <= val <= 255):
                            return False
                    except ValueError:
                        return False
            return True
        else:
            return False

    if '*' in ip and ':' in ip:
        ipv6_pattern = r'^([0-9a-fA-F]{1,4}|\*)(:([0-9a-fA-F]{1,4}|\*)){0,7}$'
        if re.match(ipv6_pattern, ip):
            octets = ip.split(':')
            for octet in octets:
                if octet != '*':
                    try:
                        val = int(octet, 16)
                        if not (0 <= val <= 0xFFFF):
                            return False
                    except ValueError:
                        return False
            return True
        else:
            return False

    try:
        ipaddress.ip_address(ip)
        return True
    except ValueError:
        return False

def parse_size(size):
    """
    Parse a string representing a size (like "5 kb" or "2.5Gb") and
    return the size in bytes.
    :param size: The size to parse
    :type size: str:
    :return int:
    :raise ValueError: if the string cannot be parsed.
    """
    try:
        val,unit = re.fullmatch(SIZE_PATTERN, str(size), flags=re.IGNORECASE).groups()
        return round(float(val) * SIZE_UNITS[unit.lower()])
    except AttributeError:
        raise ValueError(f'Unable to parse "{size}" as a size.')


def format_size(size):
    """
    Return a string representing a size (like "5 kb" or "2.5Gb")
    :param size: The size int bytes to format
    :type size: int:
    :return str:
    """
    for unit in ["B", "KB", "MB", "GB",]:
        if (size < 1024):
            return f"{size:3.1f} {unit}"
        size /= 1024
    return f"{size:.1f} TB"


def get_default_db_lib():
    """
    Get the default value for the database implementation
    :return str:  Should be either 'bdb' or 'mdb'
    """
    return os.getenv('NSSLAPD_DB_LIB', default=DEFAULT_DB_LIB)


def get_default_mdb_max_size(paths):
    """
    Get the default maximum size for the lmdb database.
    :return str:  A size that can be parsed with parse_size()
    """
    if paths is None:
        paths = Paths()
    mdb_max_size = format_size(parse_size(DEFAULT_LMDB_SIZE))
    size = parse_size(mdb_max_size)
    # Make sure that there is enough available disk space
    # otherwise decrease the value
    dbdir = paths.db_dir
    # dbdir may not exists because:
    #  - paths instance name has not been expanded
    #  - dscreate has not yet created it.
    # so let use the first existing parent in that case.
    while not os.path.exists(dbdir):
        dbdir = os.path.dirname(dbdir)
    try:
        statvfs = os.statvfs(dbdir)
        avail = statvfs.f_frsize * statvfs.f_bavail
        avail *= 0.8 # Reserve 20% as margin
        if size > avail:
            mdb_max_size = format_size(avail)
    except (TimeoutError, InterruptedError) as e:
        raise e
    except OSError as e:
        log.warning(f'Cannot determine the free space in the file system containing {dbdir} because of {e}')
    return mdb_max_size


def is_fips():
    if os.path.exists('/proc/sys/crypto/fips_enabled'):
        with open('/proc/sys/crypto/fips_enabled', 'r') as f:
            state = f.readline().strip()
            if state == '1':
                return True
            else:
                return False
    else:
        return False


def convert_timestamp(timestamp):
    """Convert createtimestamp to ctime: 20170405184656Z ----> Wed Apr  5 19:46:56 2017
    :param timestamp - A timestamp from the server
    :return - the ctime of the timestamp
    """
    if timestamp:
        time_tuple = (int(timestamp[:4]), int(timestamp[4:6]), int(timestamp[6:8]),
                      int(timestamp[8:10]), int(timestamp[10:12]), int(timestamp[12:14]),
                      0, 0, 0)
        secs = time.mktime(time_tuple)
        return time.ctime(secs)
    else:
        return ""


def elapsed_time(timestamp1, timestamp2):
    """Take two timestamps from the server and get the diff
    """
    if timestamp1 and timestamp2:
        time1_tuple = (int(timestamp1[:4]), int(timestamp1[4:6]), int(timestamp1[6:8]),
                       int(timestamp1[8:10]), int(timestamp1[10:12]), int(timestamp1[12:14]),
                       0, 0, 0)
        time2_tuple = (int(timestamp2[:4]), int(timestamp2[4:6]), int(timestamp2[6:8]),
                       int(timestamp2[8:10]), int(timestamp2[10:12]), int(timestamp2[12:14]),
                       0, 0, 0)
        time1_secs = time.mktime(time1_tuple)
        time2_secs = time.mktime(time2_tuple)
        elapsed_secs = time2_secs - time1_secs
        return str(timedelta(seconds=elapsed_secs))
    else:
        return ""


def get_task_status(inst, log, taskObj, dn=None, show_log=False, watch=False, use_json=False):
    """Get the status of a Task, if "watching" keep looping over the tasks
    :param inst: DirSrv Object
    :type inst: Object
    :param log: Logging Object
    :type log: Object
    :param taskObj: Lib389 Object (task object from plugins.py)
    :type taskObj: Object
    :param dn: A task DN to get the status for (if omitted all tasks for checked)
    :type dn: str
    :param show_log: Show the task's log attribute (could be lengthy)
    :type show_log: boolean
    :param watch: Keep looping over the task until it completes
    :type watch: boolean
    :param use_json: Report the status in JSON
    :type use_json: boolean
    """

    tasks = taskObj(inst).list()
    all_finished = False
    while not all_finished:
        task_list = []
        all_finished = True
        for task in tasks:
            task_base_dn = task.get_attr_val_utf8_l('basedn')
            if task_base_dn is None:
                # Try linked attr, it uses a different attr name
                task_base_dn = task.get_attr_val_utf8_l('linkdn')
            task_filter = task.get_attr_val_utf8_l('filter')
            status = task.get_attr_val_utf8('nsTaskStatus')
            exitcode = task.get_attr_val_utf8('nsTaskExitcode')
            task_log =  task.get_attr_val_utf8('nsTaskLog')
            task_created = task.get_attr_val_utf8('nsTaskCreated')
            task_ended = task.get_attr_val_utf8('modifyTimestamp')
            if task_base_dn is None:
                task_base_dn = ""
            if task_filter is None:
                task_filter = ""
            if status is None:
                status = ""
            if exitcode is None:
                all_finished = False
                task_ended = ""
                exitcode = "Not finished ..."
                elapsed_time_str = ""
            else:
                # Task is finished, use start and end times to calc elapsed time
                elapsed_time_str = elapsed_time(task_created, task_ended)

            if not show_log:
                task_log = ""

            if (dn is not None and dn.lower() == task.dn.lower()) or dn is None:
                    task_list.append({
                        'dn': task.dn,
                        'basedn': task_base_dn,
                        'filter': task_filter,
                        'log': task_log,
                        'status': status,
                        'created': task_created,
                        'created_str': convert_timestamp(task_created),
                        'ended': task_ended,
                        'ended_str': convert_timestamp(task_ended),
                        'elapsed_time':  elapsed_time_str,
                        'exitcode': exitcode
                    })

        # Display the status
        if use_json:
            log.info(json.dumps({"type": "list", "items": task_list}, indent=4))
        else:
            num_of_tasks = len(task_list)
            count = 1
            plural = ""
            if num_of_tasks != 1:
                plural = "s"
            log.info(f'Found {num_of_tasks} fix-up task{plural}\n')

            for task in task_list:
                log.info(f"[{count}] Task: {task['dn']}")
                log.info("-" * 80)
                log.info(f" - Base DN:       {task['basedn']}")
                if task['filter']:
                    log.info(f" - Filter:        {task['filter']}")
                log.info(f" - Status:        {task['status']}")
                log.info(f" - Started:       {task['created_str']} ({task['created']})")
                if task['ended_str']:
                    log.info(f" - Ended:         {task['ended_str']} ({task['ended']})")
                else:
                    log.info(" - Ended:")
                log.info(f" - Elapsed Time:  {task['elapsed_time']}")
                log.info(f" - Exit Code:     {task['exitcode']}")
                if show_log:
                    log.info(f" - Log:\n{task['log']}")
                log.info("")
                count += 1
        if not watch or all_finished:
            break
        elif not use_json:
            log.info("Refreshing status ...")
        time.sleep(2)


def is_cert_der(file_name):
    try:
        with open(file_name, 'tr') as check_file:
            check_file.read()
            # PEM cert
            return False
    except UnicodeDecodeError:
        # DER cert
        return True


def check_cert_info(cert_file_name, search_text):
    # Use openssl to get cert details
    cmd = [
        'openssl',
        'x509',
        '-in', cert_file_name,
        '-text',
        '-noout',
    ]
    log.debug("get_cert_details cmd: %s", format_cmd_list(cmd))
    try:
        result = check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        raise ValueError(e.output.decode('utf-8').rstrip())

    cert_text = ensure_str(result)
    return search_text.lower() in cert_text.lower()


def cert_is_ca(cert_data, pkcs12_password: Optional[Union[str, bytes]] = None):
    """
    Determine if a certificate is a CA.

    Supports PEM, DER, and PKCS#12 certificates, from bytes or file paths.
    """
    # If passed bytes directly (DER,PEM)
    cert_file = None
    if isinstance(cert_data, (bytes, bytearray)):
        data = bytes(cert_data)
    else:
        cert_file = cert_data
        try:
            with open(cert_file, "rb") as f:
                data = f.read()
        except OSError as e:
            raise ValueError(f"Unable to load certificate '{cert_data}': {e}")

    try:
        # p12
        if cert_file and cert_file.lower().endswith((".p12", ".pfx")):
            try:
                privkey, cert, _ = pkcs12.load_key_and_certificates(
                    data, ensure_bytes(pkcs12_password),
                    backend=default_backend()
                )
            except Exception as e:
                raise ValueError(f"Failed to load PKCS#12 file: {cert_file}: {e}")

            if cert is None:
                raise ValueError("No certificate found in PKCS#12 container")

        # Bytes
        elif cert_file is None:
            try:
                cert = x509.load_der_x509_certificate(data, default_backend())
            except Exception:
                cert = x509.load_pem_x509_certificate(data, default_backend())

        # File
        else:
            try:
                cert = x509.load_pem_x509_certificate(data, default_backend())
            except Exception:
                cert = x509.load_der_x509_certificate(data, default_backend())

    except ValueError as ve:
        raise ValueError(f"Unable to load certificate '{cert_file}': {ve}")

    try:
        # Check key usage
        key_usage = cert.extensions.get_extension_for_oid(x509.oid.ExtensionOID.KEY_USAGE)
        if not key_usage.value.key_cert_sign:
            return False

        # Check constraints
        basic_constraints = cert.extensions.get_extension_for_oid(
            x509.oid.ExtensionOID.BASIC_CONSTRAINTS
        )
        if not basic_constraints.value.ca:
            return False
        else:
            return True

    except x509.ExtensionNotFound:
        # No extensions, check the cert info directly
        return check_cert_info(cert_file, "CA:TRUE")

def pem_to_der(blob: bytes):
    """
    Convert PEM certificate bytes to DER format.

    :param blob: PEM encoded certificate bytes
    :return: DER encoded certificate bytes
    """
    cert = x509.load_pem_x509_certificate(blob)
    return cert.public_bytes(serialization.Encoding.DER)

def is_pem_cert(blob: bytes):
    """
    Check if the given blob is a PEM certificate.

    :param blob: Certificate data bytes
    :return: True if PEM format, else False
    """
    return b"-----BEGIN CERTIFICATE-----" in blob

def get_passwd_from_file(passwd_file):
    if os.path.exists(passwd_file):
        with open(passwd_file, 'r') as f:
            passwd = f.readline().strip()
            return passwd
    raise ValueError(f"The password file '{passwd_file}' does not exist, or can not be read.")


def find_plugin_path(plugin_name):
    p = Paths()
    plugin = None
    for ppath in glob.glob(f'{p.plugin_dir}/{plugin_name}.*'):
        if not ppath.endswith('.la'):
            plugin = ppath
    return plugin


def check_plugin_strings(plugin_name, tested_strings):
    """
    Returns a dict mapping each tested symbol to True, False or None
    """
    plugin = find_plugin_path(plugin_name)
    if plugin:
        # Otherwise looks directly for the string in the plugin
        with open(plugin, "rb") as f:
            with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mm:
                return { astring:(mm.find(astring.encode()) >=0) for astring in tested_strings }
    return { astring:None for astring in tested_strings }


def get_timeout_scale():
    """
    Get the timeout scale factor
    :return float:
    """
    scale_factor = 1.0
    if Paths().asan_enabled:
        scale_factor = 2.0

    env_value = os.getenv('DS_TIMEOUT_SCALE', default=str(scale_factor))

    try:
        return float(env_value)
    except ValueError:
        log.error(f"DS_TIMEOUT_SCALE should be a valid float. Using default value: {scale_factor}")
        return scale_factor


def rpm_is_older(pkg, version):
    """Check if an RPM package version is older than specified version"""
    # rpm module is not installed in build environment so let import it only when used.
    import rpm
    ts = rpm.TransactionSet()
    mi = ts.dbMatch('name', pkg)
    for h in mi:
        log.debug(f"{pkg} {h['version']} {version}")
        for n1,n2 in itertools.zip_longest(h['version'].split('.'), version.split('.'), fillvalue=""):
            try:
                if int(n1) < int(n2):
                    return True
            except ValueError:
                if n1 < n2:
                    return True
                if n1 > n2:
                    return False
    return False


# Validate the max-age settings for changelogs.
# The value must be a number followed by a duration unit [sSmMhHdDwW].
# The first digit can not be a zero.
def validate_max_age(value, ignore_value=None):
    if ignore_value is not None and value == ignore_value:
        return

    # check value using digits except for the last character which must be a duration unit
    if not re.match(r'^\d+[sSmMhHdDwW]$', value) or value[0] == '0':
        raise ValueError(f"Invalid max age value: {value}")


def get_mount_point(dir):
    path = os.path.realpath(os.path.abspath(dir))
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path


def get_disk_space(path):
    stats = os.statvfs(path)
    block_size = stats.f_frsize
    total_size = stats.f_blocks * block_size
    if total_size == 0:
        return {
            'total': 0,
            'available': 0,
            'used': 0,
            'percent': 0
        }

    available_size = stats.f_bavail * block_size
    used_size = total_size - available_size
    used_percent = (used_size / total_size) * 100
    return {
        'total': total_size,
        'available': available_size,
        'used': used_size,
        'percent': used_percent
    }


def check_asan_report(inst, err_string):
    """
    Check if the ASAN report contains the string.

    This is used for CI tests that expect an ASAN report to exist after
    stopping the server.

    :param inst: Instance object
    :param err_string: Error string to search for
    :return: True if the error string is found, False otherwise
    :raises ValueError: If the ASAN report does not exist
    """
    if not inst.has_asan():
        return False

    asan_report = os.path.join(inst.ds_paths.run_dir,
                               f"ns-slapd-{inst.serverid}.asan.{inst.get_pid()}")

    # Now we have the pid (and the proper file name) we can stop the server
    inst.stop()

    # Check the ASAN report
    try:
        with open(asan_report, 'r') as f:
            content = f.read()
    except (FileNotFoundError, PermissionError, OSError) as e:
        raise ValueError(f"Failed to read ASAN report '{asan_report}': {e}")

    if err_string in content:
        return True

    return False



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2018 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#


import logging
import pytest
import os
from lib389._constants import *
from lib389.schema import Schema
from lib389.tests.topologies import topology_st as topo

DEBUGGING = os.getenv("DEBUGGING", default=False)
if DEBUGGING:
    logging.getLogger(__name__).setLevel(logging.DEBUG)
else:
    logging.getLogger(__name__).setLevel(logging.INFO)
log = logging.getLogger(__name__)


def test_schema(topo):
    """Basic schema querying test

    :id: 995acc60-243b-45b0-9c1c-12bbe6a2882f
    :setup: Standalone Instance
    :steps:
        1. Get attribute info for 'uid'
        2. Check the attribute is found in the expected objectclasses (must, may)
        3. Check that the 'account' objectclass is found
    :expectedresults:
        1. Success
        2. Success
        3. Success
    """
    must_expect = ['uidObject', 'account', 'posixAccount', 'shadowAccount', 'ipaUser',
                   'sambaSamAccount']
    may_expect = ['cosDefinition', 'inetOrgPerson', 'inetUser',
                  'mailRecipient', 'nsOrgPerson', 'ipaUserOverride']
    attrtype, must, may = topo.standalone.schema.query_attributetype('uid')
    assert attrtype.names == ('uid', 'userid')
    for oc in must:
        assert oc.names[0] in must_expect
    for oc in may:
        assert oc.names[0] in may_expect
    assert topo.standalone.schema.query_objectclass('account').names == \
        ('account', )


def test_schema_reload(topo):
    """Run a schema reload task

    :id: 995acc60-243b-45b0-9c1c-12bbe6a2882e
    :setup: Standalone Instance
    :steps:
        1. Add schema reload task
        2. Schema reload task succeeds
    :expectedresults:
        1. Success
        2. Success
    """

    schema = Schema(topo.standalone)
    task = schema.reload()
    assert task.exists()
    task.wait()
    assert task.get_exit_code() == 0


def test_x_origin(topo):
    """ Test that the various forms of X-ORIGIN are handled correctly

    :id: 995acc60-243b-45b0-9c1c-12bbe6a2882d
    :setup: Standalone Instance
    :steps:
        1. Create schema file with specific/unique formats for X-ORIGIN
        2. Reload the schema file (schema reload task)
        3. List all attributes without error
        4. Confirm the expected results
    :expectedresults:
        1. Success
        2. Success
        3. Success
        4. Success
    """

    # Create a custom schema file so we can tests specific formats
    schema_file_name = topo.standalone.schemadir + '/98test.ldif'
    schema_file = open(schema_file_name, "w")
    testschema = ("dn: cn=schema\n" +
                  "attributetypes: ( 8.9.10.11.12.13.16 NAME 'testattr' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 X-ORIGIN 'USER_DEFINED' )\n" +
                  "attributetypes: ( 8.9.10.11.12.13.17 NAME 'testattrtwo' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 X-ORIGIN ( 'USER_DEFINED' 'should be not ignored!!' ) )\n")
    schema_file.write(testschema)
    schema_file.close()

    # Reload the schema
    myschema = Schema(topo.standalone)
    task = myschema.reload()
    assert task.exists()
    task.wait()
    assert task.get_exit_code() == 0

    # Now search attrs and make sure there are no errors
    myschema.get_attributetypes()
    myschema.get_objectclasses()

    # Check we have X-ORIGIN as expected
    assert " 'USER_DEFINED' " in str(myschema.query_attributetype("testattr"))
    assert " 'USER_DEFINED' " in str(myschema.query_attributetype("testattrtwo"))


if __name__ == '__main__':
    # Run isolated
    # -s for DEBUG mode
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main(["-s", CURRENT_FILE])




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

# Probably need some helpers for argparse and instance selection

from argparse import ArgumentParser
from getpass import getpass
from lib389 import DirSrv
from lib389._constants import *


class CliTool(object):
    def __init__(self, args=None):
        if args is not None:
            self.args = args
            self.ds = DirSrv(verbose=args.verbose)
        else:
            self.ds = DirSrv()

    def populate_instance_dict(self, instance):
        insts = self.ds.list(serverid=instance)
        if len(insts) != 1:
            # Raise an exception here?
            self.inst = None
            raise ValueError("No such instance %s" % instance)
        else:
            self.inst = insts[0]

    def get_rootdn_pass(self):
        if self.args.binddn is None:
            binddn = self.inst[SER_ROOT_DN]
        else:
            binddn = self.args.binddn
        # There is a dict get key thing somewhere ...
        if self.inst.get(SER_ROOT_PW, None) is None:
            prompt_txt = ('Enter password for %s on instance %s: ' %
                          (binddn,
                           self.inst[SER_SERVERID_PROP]))
            self.inst[SER_ROOT_PW] = getpass(prompt_txt)
            print("")
        return

    def connect(self):
        # Can we attempt the autobind?
        # This should be a bit cleaner perhaps
        # Perhaps an argument to the cli?
        self.ds.allocate(self.inst)
        if not self.ds.can_autobind():
            self.get_rootdn_pass()
            self.ds.allocate(self.inst)
        self.ds.open()

    def disconnect(self):
        # Is there a ds unbind / disconnect?
        self.ds.close()


def _clitools_parser():
    parser = ArgumentParser()
    parser.add_argument('-Z', '--instance',
                        help='The name of the DS instance to connect to and ' +
                             'work upon.', required=True)
    parser.add_argument('-D', '--binddn',
                        help='The bind dn to use for operations. Defaults to ' +
                             'rooddn', default=None)
    parser.add_argument('-v', '--verbose', help="Display verbose debug information", action='store_true', default=False)
    return parser

clitools_parser = _clitools_parser()

if __name__ == '__main__':
    args = clitools_parser.parse_args()



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2022 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import logging

logging.basicConfig(level=logging.DEBUG)

DN_RMANAGER = 'uid=rmanager,cn=config'

auth = {'host': 'localhost',
        'port': 22389,
        'binddn': 'cn=directory manager',
        'bindpw': 'password'}


class MockDirSrv(object):
    host = 'localhost'
    port = 22389
    sslport = 0

    def __str__(self):
        if self.sslport:
            return 'ldaps://%s:%s' % (self.host, self.sslport)
        else:
            return 'ldap://%s:%s' % (self.host, self.port)


def expect(entry, name, value):
    assert entry, "Bad entry %r " % entry
    assert entry.getValue(name) == value, \
        ("Bad value for entry %s. Expected %r vs %r" %
         (entry, entry.getValue(name), value))


def entry_equals(e1, e2):
    """compare using str()"""
    return str(e1) == str(e2)


def dfilter(my_dict, keys):
    """Filter a dict in a 2.4-compatible way"""
    return dict([(k, v) for k, v in my_dict.items() if k in keys])



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2022 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import os
import sys
import time
import ldap
import logging
import pytest
from lib389 import DirSrv, Entry, tools, tasks
from lib389.tools import DirSrvTools
from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
from lib389.utils import *

from lib389.idm.user import UserAccounts, TEST_USER_PROPERTIES
from lib389.idm.group import Groups

from lib389.tests.topologies import topology_st as topology

DEBUGGING = os.getenv('DEBUGGING', False)

if DEBUGGING is not False:
    DEBUGGING = True

if DEBUGGING:
    logging.getLogger(__name__).setLevel(logging.DEBUG)
else:
    logging.getLogger(__name__).setLevel(logging.INFO)

log = logging.getLogger(__name__)


def test_users_and_groups(topology):
    """
    Ensure that user and group management works as expected.
    """
    if DEBUGGING:
        # Add debugging steps(if any)...
        pass

    users = UserAccounts(topology.standalone, DEFAULT_SUFFIX)
    groups = Groups(topology.standalone, DEFAULT_SUFFIX)

    # No user should exist

    assert(len(users.list()) == 0)

    # Create a user

    users.create(properties=TEST_USER_PROPERTIES)

    assert(len(users.list()) == 1)

    testuser = users.get('testuser')

    # Set password
    testuser.set('userPassword', 'password')
    # bind

    conn = testuser.bind('password')
    conn.unbind_s()

    # create group
    group_properties = {
        'cn' : 'group1',
        'description' : 'testgroup'
    }

    group = groups.create(properties=group_properties)

    # user shouldn't be a member
    assert(not group.is_member(testuser.dn))

    # add user as member
    group.add_member(testuser.dn)
    # check they are a member
    assert(group.is_member(testuser.dn))
    # remove user from group
    group.remove_member(testuser.dn)
    # check they are not a member
    assert(not group.is_member(testuser.dn))

    group.delete()
    testuser.delete()

    log.info('Test PASSED')


if __name__ == '__main__':
    # Run isolated
    # -s for DEBUG mode
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s %s" % CURRENT_FILE)




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2025 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
from lib389.config import Config
from lib389.cli_base import CustomHelpFormatter

ERROR_LEVELS = {
    "acl": {
        "level": 128,
        "desc":
            "Provides very detailed access control list processing information"
    },
    "aclsumary": {
        "level": 262144,
        "desc": "Summarizes information about access to the server, much less "
                + "verbose than level 'acl'"
    },
    "backend": {
        "level": 524288,
        "desc": "Backend debug logging"
    },
    "ber": {
        "level": 16,
        "desc": "Logs the number of packets sent and received by the server"
    },
    "cache": {
        "level": 32768,
        "desc": "Database entry cache logging"
    },
    "config": {
        "level": 64,
        "desc": "Prints any .conf configuration files used with the server, "
                + "line by line, when the server is started"
    },
    "connection": {
        "level": 8,
        "desc": "Logs the current connection status, including the connection "
                + "methods used for a SASL bind"
    },
    "default": {
        "level": 16384,
        "desc": "Default logging level"
    },
    "filter": {
        "level": 32,
        "desc": "Logs all of the functions called by a search operation"
    },
    "heavytrace": {
        "level": 4,
        "desc": "Logs when the server enters and exits a function, with "
                + "additional debugging messages"
    },
    "house": {
        "level": 4096,
        "desc": "Logging for housekeeping thread"
    },
    "packet":
    {
        "level": 2,
        "desc": "Network packet logging"
    },
    "parse": {
        "level": 2048,
        "desc": "Logs schema parsing debugging information"
    },
    "plugin": {
        "level": 65536,
        "desc": "Plugin logging"
    },
    "pwpolicy": {
        "level": 1048576,
        "desc": "Debug information about password policy behavior"
    },
    "replication": {
        "level": 8192,
        "desc": "Debug replication logging"
    },
    "shell": {
        "level": 1024,
        "desc": "Special authentication/connection tracking"
    },
    "trace": {
        "level": 1,
        "desc": "Logs a message when the server enters and exits a function"
    },
}

ACCESS_LEVELS = {
    "entry": {
        "level": 512,
        "desc": "Log entry and referral stats"
    },
    "default": {
        "level": 256,
        "desc": "Standard access logging"
    },
    "internal": {
        "level": 4,
        "desc": "Log internal operations"
    },
}

ACCESS_ATTR_MAP = {
    'nsslapd-accesslog': 'Log name and location',
    'nsslapd-accesslog-mode': 'File mode',
    'nsslapd-accesslog-maxlogsperdir': 'Max logs',
    'nsslapd-accesslog-logging-enabled': 'Logging enabled',
    'nsslapd-accesslog-compress': 'Compression enabled',
    'nsslapd-accesslog-logexpirationtime': 'Deletion interval',
    'nsslapd-accesslog-logexpirationtimeunit': 'Deletion interval unit',
    'nsslapd-accesslog-logrotationsync-enabled': 'TOD rotation enabled',
    'nsslapd-accesslog-logrotationsynchour': 'TOD rotation hour',
    'nsslapd-accesslog-logrotationsyncmin': 'TOD rotation minute',
    'nsslapd-accesslog-logrotationtime': 'Rotation interval',
    'nsslapd-accesslog-logrotationtimeunit': 'Rotation interval unit',
    'nsslapd-accesslog-logmaxdiskspace': 'Max disk space',
    'nsslapd-accesslog-level': 'Log level',
    'nsslapd-accesslog-maxlogsize': 'Max log size',
    'nsslapd-accesslog-logbuffering': 'Buffering enabled',
    'nsslapd-accesslog-logminfreediskspace': 'Minimum free disk space',
    'nsslapd-accesslog-time-format': 'Time format for JSON logging (strftime)',
    'nsslapd-accesslog-log-format': 'Logging format',
}

AUDIT_ATTR_MAP = {
    'nsslapd-auditlog': 'Log name and location',
    'nsslapd-auditlog-mode': 'File mode',
    'nsslapd-auditlog-maxlogsperdir': 'Max logs',
    'nsslapd-auditlog-logging-enabled': 'Logging enabled',
    'nsslapd-auditlog-compress': 'Compression enabled',
    'nsslapd-auditlog-logexpirationtime': 'Deletion interval',
    'nsslapd-auditlog-logexpirationtimeunit': 'Deletion interval unit',
    'nsslapd-auditlog-logrotationsync-enabled': 'TOD rotation enabled',
    'nsslapd-auditlog-logrotationsynchour': 'TOD rotation hour',
    'nsslapd-auditlog-logrotationsyncmin': 'TOD rotation minute',
    'nsslapd-auditlog-logrotationtime': 'Rotation interval',
    'nsslapd-auditlog-logrotationtimeunit': 'Rotation interval unit',
    'nsslapd-auditlog-logmaxdiskspace': 'Max disk space',
    'nsslapd-auditlog-maxlogsize': 'Max log size',
    'nsslapd-auditlog-logbuffering': 'Buffering enabled',
    'nsslapd-auditlog-logminfreediskspace': 'Minimum free disk space',
    'nsslapd-auditlog-display-attrs': 'Additional attrs to display',
    'nsslapd-auditlog-time-format': 'Time format for JSON logging (strftime)',
    'nsslapd-auditlog-log-format': 'Logging format',
}

AUDITFAIL_ATTR_MAP = {
    'nsslapd-auditfaillog': 'Log name and location',
    'nsslapd-auditfaillog-mode': 'File mode',
    'nsslapd-auditfaillog-maxlogsperdir': 'Max logs',
    'nsslapd-auditfaillog-logging-enabled': 'Logging enabled',
    'nsslapd-auditfaillog-compress': 'Compression enabled',
    'nsslapd-auditfaillog-logexpirationtime': 'Deletion interval',
    'nsslapd-auditfaillog-logexpirationtimeunit': 'Deletion interval unit',
    'nsslapd-auditfaillog-logrotationsync-enabled': 'TOD rotation enabled',
    'nsslapd-auditfaillog-logrotationsynchour': 'TOD rotation hour',
    'nsslapd-auditfaillog-logrotationsyncmin': 'TOD rotation minute',
    'nsslapd-auditfaillog-logrotationtime': 'Rotation interval',
    'nsslapd-auditfaillog-logrotationtimeunit': 'Rotation interval unit',
    'nsslapd-auditfaillog-logmaxdiskspace': 'Max disk space',
    'nsslapd-auditfaillog-maxlogsize': 'Max log size',
    'nsslapd-auditfaillog-logminfreediskspace': 'Minimum free disk space',
}

ERROR_ATTR_MAP = {
    'nsslapd-errorlog': 'Log name and location',
    'nsslapd-errorlog-mode': 'File mode',
    'nsslapd-errorlog-maxlogsperdir': 'Max logs',
    'nsslapd-errorlog-logging-enabled': 'Logging enabled',
    'nsslapd-errorlog-compress': 'Compression enabled',
    'nsslapd-errorlog-logexpirationtime': 'Deletion interval',
    'nsslapd-errorlog-logexpirationtimeunit': 'Deletion interval unit',
    'nsslapd-errorlog-logrotationsync-enabled': 'TOD rotation enabled',
    'nsslapd-errorlog-logrotationsynchour': 'TOD rotation hour',
    'nsslapd-errorlog-logrotationsyncmin': 'TOD rotation minute',
    'nsslapd-errorlog-logrotationtime': 'Rotation interval',
    'nsslapd-errorlog-logrotationtimeunit': 'Rotation interval unit',
    'nsslapd-errorlog-logmaxdiskspace': 'Max disk space',
    'nsslapd-errorlog-level': 'Log level',
    'nsslapd-errorlog-maxlogsize': 'Max log size',
    'nsslapd-errorlog-logbuffering': 'Buffering enabled',
    'nsslapd-errorlog-logminfreediskspace': 'Minimum free disk space',
    'nsslapd-errorlog-time-format': 'Time format for JSON logging (strftime)',
    'nsslapd-errorlog-log-format': 'Logging format',
}

SECURITY_ATTR_MAP = {
    'nsslapd-securitylog': 'Log name and location',
    'nsslapd-securitylog-mode': 'File mode',
    'nsslapd-securitylog-maxlogsperdir': 'Max logs',
    'nsslapd-securitylog-logging-enabled': 'Logging enabled',
    'nsslapd-securitylog-compress': 'Compression enabled',
    'nsslapd-securitylog-logexpirationtime': 'Deletion interval',
    'nsslapd-securitylog-logexpirationtimeunit': 'Deletion interval unit',
    'nsslapd-securitylog-logrotationsync-enabled': 'TOD rotation enabled',
    'nsslapd-securitylog-logrotationsynchour': 'TOD rotation hour',
    'nsslapd-securitylog-logrotationsyncmin': 'TOD rotation minute',
    'nsslapd-securitylog-logrotationtime': 'Rotation interval',
    'nsslapd-securitylog-logrotationtimeunit': 'Rotation interval unit',
    'nsslapd-securitylog-logmaxdiskspace': 'Max disk space',
    'nsslapd-securitylog-maxlogsize': 'Max log size',
    'nsslapd-securitylog-logbuffering': 'Buffering enabled',
    'nsslapd-securitylog-logminfreediskspace': 'Minimum free disk space',
}


###############################################################################
# Generic functions
###############################################################################

def update_config(inst, basedn, log, args):
    """
    Update a simple log config attribute
    """
    base = "nsslapd-" + args.logtype + "log"
    if 'keyword' in args and args.keyword is not None:
        attr = base + "-" + args.keyword
    else:
        attr = base
    config = Config(inst, basedn)
    config.set(attr, args.values)
    log.info(f"Successfully updated {args.logtype} log configuration")


def get_log_config(inst, basedn, log, args):
    """
    Get all the configuration settings for the log
    """
    conf = Config(inst, basedn)
    attrs = conf.get_all_attrs_utf8()
    json_result = {}
    attr_map = {}
    levels = {}

    if args.logtype == "access":
        attr_map = ACCESS_ATTR_MAP
        levels = ACCESS_LEVELS
    elif args.logtype == "error":
        attr_map = ERROR_ATTR_MAP
        levels = ERROR_LEVELS
    elif args.logtype == "security":
        attr_map = SECURITY_ATTR_MAP
    elif args.logtype == "audit":
        attr_map = AUDIT_ATTR_MAP
    elif args.logtype == "auditfail":
        attr_map = AUDITFAIL_ATTR_MAP
    else:
        raise ValueError(f"Unknown logtype: {args.logtype}")

    sorted_results = []
    for attr, value in attrs.items():
        attr = attr.lower()
        if attr in attr_map:
            attr_val = value[0]
            if 'level' in attr:
                # Convert log level number into individual levels
                level_vals = []
                for name, obj in levels.items():
                    if int(attr_val) & obj['level']:
                        level_vals.append(name)
                level_vals.sort()
                if args.json:
                    attr_val = level_vals
                else:
                    attr_val = (',').join(level_vals)
            elif 'display-attrs' in attr:
                display_attr_vals = value[0].split()
                display_attr_vals.sort()
                if args.json:
                    attr_val = display_attr_vals
                else:
                    attr_val = (',').join(display_attr_vals)

            if args.json:
                json_result[attr_map[attr]] = attr_val
            else:
                sorted_results.append(attr_map[attr] + " = " + attr_val)

    if args.json:
        log.info(json.dumps(json_result, indent=4))
    else:
        sorted_results.sort()
        header = args.logtype + " log configuration"
        log.info("\n" + header.title())
        log.info('-' * 80)
        for result in sorted_results:
            log.info(result)


def list_log_levels(inst, basedn, log, args):
    """
    List all the log levels
    """
    if args.logtype == 'access':
        levels = ACCESS_LEVELS
    else:  # args.logtype == 'error'
        levels = ERROR_LEVELS

    result = []
    if not args.json:
        log.info('\nLevel         Description')
        log.info('-' * 80)

    for name, obj in levels.items():
        desc = obj['desc']
        if args.json:
            result.append({name: desc})
        else:
            space = 14 - len(name)
            log.info(name + (" " * space) + desc)

    if args.json:
        result = {'type': 'list', 'items': result}
        log.info(json.dumps(result, indent=4))


def set_log_level(inst, basedn, log, args):
    """
    Set the log level
    """
    new_level = 0
    attr = f'nsslapd-{args.logtype}log-level'

    if args.logtype == 'access':
        level_map = ACCESS_LEVELS
    else:  # args.logtype == 'error'
        level_map = ERROR_LEVELS

    for level in args.levels:
        if level in level_map:
            new_level += level_map[level]['level']
        else:
            log.error('Ignoring unknown log level: ' + level)
    if new_level > 0:
        config = Config(inst, basedn)
        config.set(attr, str(new_level))
        log.info(f'Successfully set {args.logtype} log level')


def set_display_attrs(inst, basedn, log, args):
    """
    Set the log level
    """
    attr = 'nsslapd-auditlog-display-attrs'
    value = (" ").join(args.values)
    value = value.replace(",", " ")
    config = Config(inst, basedn)
    config.set(attr, value)
    log.info('Successfully updated audit log configuration')


#
# CLI parser
#
def create_parser(subparsers):
    """
    CLI log parser
    """
    log_parser = subparsers.add_parser(
        "logging", help="Manage the server logs",
        formatter_class=CustomHelpFormatter
    )
    log_sub_parsers = log_parser.add_subparsers()

    for log_type in ['access', 'audit', 'auditfail', 'error', 'security']:
        subcommands = log_sub_parsers.add_parser(
            log_type, help="Manage " + log_type + " log settings",
            formatter_class=CustomHelpFormatter
        )
        log_subparsers = subcommands.add_subparsers()

        # Get config
        get_parser = log_subparsers.add_parser(
            "get", help="Get " + log_type + " log configuration",
            formatter_class=CustomHelpFormatter
        )
        get_parser.set_defaults(func=get_log_config, logtype=log_type)

        # Set config
        set_parser = log_subparsers.add_parser(
            "set", help="Set " + log_type + " log configuration",
            formatter_class=CustomHelpFormatter
        )
        set_parsers = set_parser.add_subparsers()

        # Log levels
        if log_type in ['access', 'error']:
            # List levels
            list_parser = log_subparsers.add_parser(
                "list-levels",
                help="List all the log levels",
                formatter_class=CustomHelpFormatter,
            )
            list_parser.set_defaults(func=list_log_levels, logtype=log_type)

            # Set level
            set_level_parser = set_parsers.add_parser(
                "level", help="Set the log level",
                formatter_class=CustomHelpFormatter
            )
            set_level_parser.set_defaults(func=set_log_level, logtype=log_type)
            set_level_parser.add_argument(
                "levels", nargs="+", help="log level"
            )

        # Enable disable logging
        on_parser = set_parsers.add_parser(
            "logging-enabled",
            help="Enable access logging",
            formatter_class=CustomHelpFormatter,
        )
        on_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logging-enabled",
            values=["on"]
        )

        off_parser = set_parsers.add_parser(
            "logging-disabled",
            help="Disable " + log_type + " logging",
            formatter_class=CustomHelpFormatter,
        )
        off_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logging-enabled",
            values=["off"]
        )

        # Mode/permissions
        set_mode_parser = set_parsers.add_parser(
            "mode",
            help="Set the log file permissions. Default is 600",
            formatter_class=CustomHelpFormatter,
        )
        set_mode_parser.set_defaults(func=update_config, logtype=log_type,
                                     keyword="mode")
        set_mode_parser.add_argument(
            "values", nargs=1,
            help="File permissions. Default is 600"
        )

        # Location
        loc_set_parser = set_parsers.add_parser(
            "location",
            help="Set the log name and location",
            formatter_class=CustomHelpFormatter,
        )
        loc_set_parser.set_defaults(func=update_config, logtype=log_type)
        loc_set_parser.add_argument(
            "values", nargs=1, help="Log name and location"
        )

        # Compression
        comp_on_parser = set_parsers.add_parser(
            "compress-enabled",
            help="Enable log compression for rotated logs",
            formatter_class=CustomHelpFormatter,
        )
        comp_on_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="compress",
            values=["on"]
        )

        comp_off_parser = set_parsers.add_parser(
            "compress-disabled",
            help="Disable log compression for rotated logs",
            formatter_class=CustomHelpFormatter,
        )
        comp_off_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="compress",
            values=["off"]
        )

        # Buffering
        if log_type != "auditfail":
            buf_on_parser = set_parsers.add_parser(
                "buffering-enabled",
                help="Enable log buffering",
                formatter_class=CustomHelpFormatter,
            )
            buf_on_parser.set_defaults(
                func=update_config, logtype=log_type, keyword="logbuffering",
                values=["on"]
            )

            buf_off_parser = set_parsers.add_parser(
                "buffering-disabled",
                help="Disable log buffering",
                formatter_class=CustomHelpFormatter,
            )
            buf_off_parser.set_defaults(
                func=update_config, logtype=log_type, keyword="logbuffering",
                values=["off"]
            )

        # Log rotation size settings
        set_maxlog_parser = set_parsers.add_parser(
            "max-logs",
            help="Set the maximum number of rotated logs the server "
            + "will maintain",
            formatter_class=CustomHelpFormatter,
        )
        set_maxlog_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="maxlogsperdir",
            values=["off"]
        )
        set_maxlog_parser.add_argument(
            "values",
            nargs=1,
            help="Set the maximum number of rotated logs the server "
            + "will maintain",
        )

        set_logsize_parser = set_parsers.add_parser(
            "max-logsize",
            help="Set the maximum size for a log in MB",
            formatter_class=CustomHelpFormatter,
        )
        set_logsize_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="maxlogsize"
        )
        set_logsize_parser.add_argument(
            "values", nargs=1, help="Set the maximum size for a log in MB"
        )

        # Log rotation interval settings
        set_interval_parser = set_parsers.add_parser(
            "rotation-interval",
            help="Set the interval for when a log is rotated."
            + "This works with the interval unit",
            formatter_class=CustomHelpFormatter,
        )
        set_interval_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logrotationtime"
        )
        set_interval_parser.add_argument(
            "values",
            nargs=1,
            help="Set the interval for when a log is rotated."
            + "This works with the interval unit",
        )

        set_interval_unit_parser = set_parsers.add_parser(
            "rotation-interval-unit",
            help="Set the time unit for the rotation interval for when"
            + 'a log is rotated.  Choose between: "minute", "hour", "day", '
            + '"week", and "month"',
            formatter_class=CustomHelpFormatter,
        )
        set_interval_unit_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logrotationtimeunit"
        )
        set_interval_unit_parser.add_argument(
            "values",
            nargs=1,
            help="Set the time unit for the rotation interval for when"
            + 'a log is rotated.  Choose between: "minute", "hour", "day", '
            + '"week", and "month"',
        )

        # Log rotation TOD settings
        set_tod_enabled = set_parsers.add_parser(
            "rotation-tod-enabled",
            help='Enable "time of day" rotation for expired logs',
            formatter_class=CustomHelpFormatter,
        )
        set_tod_enabled.set_defaults(
            func=update_config,
            logtype=log_type,
            keyword="logrotationsync-enabled",
            values=["on"],
        )

        set_tod_disabled = set_parsers.add_parser(
            "rotation-tod-disabled",
            help='Disable "time of day" rotation for expired logs',
            formatter_class=CustomHelpFormatter,
        )
        set_tod_disabled.set_defaults(
            func=update_config,
            logtype=log_type,
            keyword="logrotationsync-enabled",
            values=["off"],
        )

        set_tod_hour_parser = set_parsers.add_parser(
            "rotation-tod-hour",
            help="Set the hour when an expired log should be rotated",
            formatter_class=CustomHelpFormatter,
        )
        set_tod_hour_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logrotationsynchour"
        )
        set_tod_hour_parser.add_argument(
            "values", nargs=1,
            help="Set the hour when an expired log should be rotated"
        )

        set_tod_hour_parser = set_parsers.add_parser(
            "rotation-tod-minute",
            help="Set the minute when an expired log should be rotated",
            formatter_class=CustomHelpFormatter,
        )
        set_tod_hour_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logrotationsyncmin"
        )
        set_tod_hour_parser.add_argument(
            "values", nargs=1,
            help="Set the minute when an expired log should be rotated"
        )

        # Log deletion settings
        set_expire_interval_parser = set_parsers.add_parser(
            "deletion-interval",
            help="Set the interval a rotated log should be deleted. "
            + "This works with the deletion internal unit setting",
            formatter_class=CustomHelpFormatter,
        )
        set_expire_interval_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logexpirationtime"
        )
        set_expire_interval_parser.add_argument(
            "values",
            nargs=1,
            help="Set the interval a rotated log should be deleted. "
            + "This works with the deletion internal unit setting",
        )

        set_expire_interval_unit_parser = set_parsers.add_parser(
            "deletion-interval-unit",
            help="Set the interval unit a rotated log should be deleted. "
            + 'Choose from: "day", "week", or "month"',
            formatter_class=CustomHelpFormatter,
        )
        set_expire_interval_unit_parser.set_defaults(
            func=update_config, logtype=log_type,
            keyword="logexpirationtimeunit"
        )
        set_expire_interval_unit_parser.add_argument(
            "values",
            nargs=1,
            help="Set the interval unit a rotated log should be deleted. "
            + 'Choose from: "day", "week", or "month"',
        )

        set_max_disk_space_parser = set_parsers.add_parser(
            "max-disk-space",
            help="Set the maximum amount of disk space in MB rotated logs can "
            + "consume before rotated logs are deleted.",
            formatter_class=CustomHelpFormatter,
        )
        set_max_disk_space_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logmaxdiskspace"
        )
        set_max_disk_space_parser.add_argument(
            "values",
            nargs=1,
            help="Set the maximum amount of disk space in MB rotated logs can "
            + "consume before rotated logs are deleted.",
        )

        set_free_disk_space_parser = set_parsers.add_parser(
            "free-disk-space",
            help="The server deletes the oldest rotated log file when the "
            + "available disk space in MB is less than this amount.",
            formatter_class=CustomHelpFormatter,
        )
        set_free_disk_space_parser.set_defaults(
            func=update_config, logtype=log_type, keyword="logminfreediskspace"
        )
        set_free_disk_space_parser.add_argument(
            "values",
            nargs=1,
            help="Set the minimum available disk space in MB that triggers "
            + "the server to delete rotated log files.",
        )

        if log_type in ['access', 'audit', 'error']:
            # JSON logging
            set_log_format_parser = set_parsers.add_parser(
                "log-format",
                help='Choose between "default", "json", or "json-pretty"',
                formatter_class=CustomHelpFormatter,
            )
            set_log_format_parser.set_defaults(
                func=update_config, logtype=log_type, keyword="log-format"
            )
            set_log_format_parser.add_argument(
                "values", nargs=1,
                help='Choose between "default", "json", or "json-pretty"'
            )

            set_time_format_parser = set_parsers.add_parser(
                "time-format",
                help="Time format for JSON logging (strftime)",
                formatter_class=CustomHelpFormatter,
            )
            set_time_format_parser.set_defaults(
                func=update_config, logtype=log_type, keyword="time-format"
            )
            set_time_format_parser.add_argument(
                "values", nargs=1, help="Time format for JSON logging "
                                        + "(strftime)"
            )

        if log_type == 'audit':
            set_display_attrs_parser = set_parsers.add_parser(
                "display-attrs",
                help="Sets additional identifying attrs to display",
                formatter_class=CustomHelpFormatter,
            )
            set_display_attrs_parser.set_defaults(func=set_display_attrs)
            set_display_attrs_parser.add_argument(
                "values", nargs="+",
                help="Sets additional identifying attrs to display"
            )



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

"""
   You will access this from:
   schema = Schema(instance)
"""
import glob
import ldap
import ldif
import re
from itertools import count
from json import dumps as dump_json
from operator import itemgetter
from ldap.schema.models import AttributeType, ObjectClass, MatchingRule
from lib389._constants import *
from lib389._constants import DN_SCHEMA
from lib389.utils import ds_is_newer
from lib389._mapped_object import DSLdapObject
from lib389.tasks import SchemaReloadTask, SyntaxValidateTask

# Count should start with 0 because of the python-ldap API
ObjectclassKind = Enum("Objectclass kind",
                       zip(["STRUCTURAL", "ABSTRACT", "AUXILIARY"], count()))
AttributeUsage = Enum("Attribute usage",
                      zip(["userApplications", "directoryOperation", "distributedOperation", "dSAOperation"], count()))

OBJECT_MODEL_PARAMS = {ObjectClass: {'names': (), 'oid': None, 'desc': None, 'x_origin': None, 'obsolete': 0,
                                     'kind': ObjectclassKind.STRUCTURAL.value, 'sup': (), 'must': (), 'may': ()},
                       AttributeType: {'names': (), 'oid': None, 'desc': None, 'sup': (), 'obsolete': 0,
                                       'equality': None, 'ordering': None, 'substr': None, 'collective': 0,
                                       'syntax': None, 'syntax_len': None, 'single_value': 0,
                                       'no_user_mod': 0, 'usage': AttributeUsage.userApplications.value,
                                       'x_origin': None, 'x_ordered': None}}
ATTR_SYNTAXES = {"1.3.6.1.4.1.1466.115.121.1.5": "Binary",
                 "1.3.6.1.4.1.1466.115.121.1.6": "Bit String",
                 "1.3.6.1.4.1.1466.115.121.1.7": "Boolean",
                 "1.3.6.1.4.1.1466.115.121.1.11": "Country String",
                 "1.3.6.1.4.1.1466.115.121.1.12": "DN",
                 "1.3.6.1.4.1.1466.115.121.1.14": "Delivery Method",
                 "1.3.6.1.4.1.1466.115.121.1.15": "Directory String",
                 "1.3.6.1.4.1.1466.115.121.1.21": "Enhanced Guide",
                 "1.3.6.1.4.1.1466.115.121.1.22": "Facsimile",
                 "1.3.6.1.4.1.1466.115.121.1.23": "Fax",
                 "1.3.6.1.4.1.1466.115.121.1.24": "Generalized Time",
                 "1.3.6.1.4.1.1466.115.121.1.25": "Guide",
                 "1.3.6.1.4.1.1466.115.121.1.26": "IA5 String",
                 "1.3.6.1.4.1.1466.115.121.1.27": "Integer",
                 "1.3.6.1.4.1.1466.115.121.1.28": "JPEG",
                 "1.3.6.1.4.1.1466.115.121.1.34": "Name and Optional UID",
                 "1.3.6.1.4.1.1466.115.121.1.36": "Numeric String",
                 "1.3.6.1.4.1.1466.115.121.1.40": "OctetString",
                 "1.3.6.1.4.1.1466.115.121.1.37": "Object Class Description",
                 "1.3.6.1.4.1.1466.115.121.1.38": "OID",
                 "1.3.6.1.4.1.1466.115.121.1.41": "Postal Address",
                 "1.3.6.1.4.1.1466.115.121.1.44": "Printable String",
                 "2.16.840.1.113730.3.7.1": "Space-Insensitive String",
                 "1.3.6.1.4.1.1466.115.121.1.50": "TelephoneNumber",
                 "1.3.6.1.4.1.1466.115.121.1.51": "Teletex Terminal Identifier",
                 "1.3.6.1.4.1.1466.115.121.1.52": "Telex Number"}


X_ORIGIN_REGEX = r'\'(.*?)\''


class Schema(DSLdapObject):
    """An object that represents the schema entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    """

    def __init__(self, instance):
        super(Schema, self).__init__(instance=instance)
        self._dn = DN_SCHEMA
        self._rdn_attribute = 'cn'

    @staticmethod
    def _get_attr_name_by_model(object_model):
        # Validate the model and return its attribute name
        if object_model in (ObjectClass, AttributeType, MatchingRule):
            return object_model.schema_attribute
        else:
            raise ValueError("Wrong object model was specified")

    @staticmethod
    def _validate_ldap_schema_value(value):
        """Validate the values that we supply to ldap.schema.models
        because it expects some exact values.
        It should be tuple, not list.
        It should be None or () if we don't want """

        if type(value) == list:
            value = tuple(value)
        elif value == "":
            value = None
        if value == ("",):
            value = ()
        return value

    @staticmethod
    def get_attr_syntaxes(json=False):
        """Get a list of available attribute syntaxes"""

        if json:
            attr_syntaxes_list = []
            for id, name in ATTR_SYNTAXES.items():
                attr_syntaxes_list.append({'label': name, 'id': id})
            result = {'type': 'list', 'items': attr_syntaxes_list}
        else:
            result = ATTR_SYNTAXES
        return result

    def gather_oc_sup_attrs(self, oc, sup_oc, ocs, processed_ocs=None):
        """
        Recursively build up all the objectclass superiors' may/must
        attributes

        @param oc - original objectclass we are building up
        @param sup_oc - superior objectclass that we are gathering must/may
                        attributes from, and for following its superior
                        objectclass
        @param ocs - all objectclasses
        @param processed_ocs - list of all the superior objectclasees we have
                               already processed. Used for checking if we
                               somehow get into an infinite loop
        """
        if processed_ocs is None:
            # First pass, init our values
            sup_oc = oc
            processed_ocs = [sup_oc['names'][0]]
        elif sup_oc['names'][0] in processed_ocs:
            # We're looping, need to abort. This should never happen because
            # of how the schema is structured, but perhaps a bug was
            # introduced in the server schema handling?
            return

        # update processed list to prevent loops
        processed_ocs.append(sup_oc['names'][0])

        for soc in sup_oc['sup']:
            if soc.lower() == "top":
                continue
            # Get sup_oc
            for obj in ocs:
                oc_dict = vars(ObjectClass(obj))
                name = oc_dict['names'][0]
                if name.lower() == soc.lower():
                    # Found the superior, get it's attributes
                    for attr in oc_dict['may']:
                        if attr not in oc['may']:
                            oc['may'] = oc['may'] + (attr,)
                    for attr in oc_dict['must']:
                        if attr not in oc['must']:
                            oc['must'] = oc['must'] + (attr,)

                    # Sort the tuples
                    oc['may'] = tuple(sorted(oc['may']))
                    oc['must'] = tuple(sorted(oc['must']))

                    # Now recurse and check this objectclass
                    self.gather_oc_sup_attrs(oc, oc_dict, ocs, processed_ocs)

    def _get_schema_objects(self, object_model, include_sup=False, json=False):
        """Get all the schema objects for a specific model:

            Attribute, ObjectClass, or MatchingRule.
        """
        attr_name = self._get_attr_name_by_model(object_model)
        results = self.get_attr_vals_utf8(attr_name)
        object_insts = []

        if json:
            for obj in results:
                obj_i = vars(object_model(obj))
                if len(obj_i["names"]) == 1:
                    obj_i['name'] = obj_i['names'][0]
                    obj_i['aliases'] = None
                elif len(obj_i["names"]) > 1:
                    obj_i['name'] = obj_i['names'][0]
                    obj_i['aliases'] = obj_i['names'][1:]
                else:
                    obj_i['name'] = ""

                if object_model is ObjectClass and include_sup:
                    self.gather_oc_sup_attrs(obj_i, None, results)

                object_insts.append(obj_i)

            object_insts = sorted(object_insts, key=itemgetter('name'))
            # Ensure that the string values are in list so we can use React filter component with it
            for obj_i in object_insts:
                for key, value in obj_i.items():
                    if isinstance(value, str):
                        obj_i[key] = (value, )

            return {'type': 'list', 'items': object_insts}
        else:
            for obj_i in results:
                obj_i = object_model(obj_i)
                if object_model is ObjectClass and include_sup:
                    obj_ii = vars(obj_i)
                    self.gather_oc_sup_attrs(obj_ii, None, results)
                    obj_i.may = obj_ii['may']
                    obj_i.must = obj_ii['must']
                object_insts.append(obj_i)
            return sorted(object_insts, key=lambda x: x.names, reverse=False)

    def _get_schema_object(self, name, object_model, include_sup=False, json=False):
        objects = self._get_schema_objects(object_model,
                                           include_sup=include_sup,
                                           json=json)
        if json:
            schema_object = [obj_i for obj_i in objects["items"] if name.lower() in
                             list(map(str.lower, obj_i["names"]))]
        else:
            schema_object = [obj_i for obj_i in objects if name.lower() in
                             list(map(str.lower, obj_i.names))]

        if len(schema_object) != 1:
            # This is an error.
            if json:
                raise ValueError('Could not find: %s' % name)
            else:
                return None

        return schema_object[0]

    def _add_schema_object(self, parameters, object_model):
        attr_name = self._get_attr_name_by_model(object_model)

        if len(parameters) == 0:
            raise ValueError('Parameters should be specified')

        # Validate args
        if "names" not in parameters.keys():
            raise ValueError('%s name should be specified' % attr_name)
        for name in parameters["names"]:
            schema_object_old = self._get_schema_object(name, object_model)
            if schema_object_old is not None:
                raise ValueError('The %s with the name %s already exists' % (attr_name, name))

        # Default structure. We modify it later with the specified arguments
        schema_object = object_model()

        # x-origin requires special handling to make sure it is a Tuple
        empty = ()
        setattr(schema_object, "x_origin", empty)

        for oc_param, value in parameters.items():
            if oc_param.lower() not in OBJECT_MODEL_PARAMS[object_model].keys():
                raise ValueError('Wrong parameter name was specified: %s' % oc_param)
            if oc_param == "x_origin" and value is not None and \
                (type(value) != list and type(value) != tuple):
                # x-origin is expected to be a tuple in python-ldap
                value = (value,)
            if value is not None:
                value = self._validate_ldap_schema_value(value)
                setattr(schema_object, oc_param.lower(), value)

        # Set other not defined arguments so objectClass model work correctly
        # all 'None', but OBSOLETE and KIND are '0' (STRUCTURAL)
        # It is automatically assigned to 'SUP top'
        parameters_none = {k.lower(): v for k, v in parameters.items() if v is None}
        for k, v in parameters_none.items():
            if k == "x_origin" and v is None:
                continue
            setattr(schema_object, k, OBJECT_MODEL_PARAMS[object_model][k])
        return self.add(attr_name, str(schema_object))

    def _remove_schema_object(self, name, object_model):
        attr_name = self._get_attr_name_by_model(object_model)
        schema_object = self._get_schema_object(name, object_model)
        return self.remove(attr_name, str(schema_object))

    def _edit_schema_object(self, name, parameters, object_model):
        attr_name = self._get_attr_name_by_model(object_model)
        schema_object = self._get_schema_object(name, object_model)
        schema_object_str_old = str(schema_object)
        superior_is_set = False
        if len(parameters) == 0:
            raise ValueError('Parameters should be specified')

        if schema_object is None:
            raise ValueError(f'Schema ({name}) not found')

        if len(schema_object.sup) > 0:
            superior_is_set = True

        for oc_param, value in parameters.items():
            if oc_param.lower() not in OBJECT_MODEL_PARAMS[object_model].keys():
                raise ValueError('Wrong parameter name was specified: %s' % oc_param)
            if value is not None:
                value = self._validate_ldap_schema_value(value)
                setattr(schema_object, oc_param, value)
            else:
                if not getattr(schema_object, oc_param, False) or not value:
                    # Need to set the correct "type" for the empty value
                    if oc_param in ['may', 'must',  'x-origin', 'sup']:
                        # Expects tuple
                        setattr(schema_object, oc_param, ())
                    elif oc_param in ['desc', 'oid']:
                        # Expects None
                        setattr(schema_object, oc_param, None)
                    elif oc_param in ['obsolete', 'kind']:
                        # Expects numberic
                        setattr(schema_object, oc_param, 0)

        if len(schema_object.sup) == 0 and superior_is_set:
            # removing a superior mean we need to remove
            # matching rules since they were inherited from sup
            setattr(schema_object, 'equality', None)
            setattr(schema_object, 'ordering', None)
            setattr(schema_object, 'substr', None)

        schema_object_str = str(schema_object)
        if schema_object_str == schema_object_str_old:
            raise ValueError('Schema is already in the required state. Nothing to change')

        self.remove(attr_name, schema_object_str_old)
        try:
            return self.add(attr_name, schema_object_str)
        except ldap.LDAPError:
            self.add(attr_name, schema_object_str_old)
            raise

    def reload(self, schema_dir=None):
        """Reload the schema"""

        task = SchemaReloadTask(self._instance)

        task_properties = {}
        if schema_dir is not None:
            task_properties['schemadir'] = schema_dir

        task.create(properties=task_properties)

        return task

    def list_files(self):
        """Return a list of the schema files in the instance schemadir"""

        file_list = []
        file_list += glob.glob(os.path.join(self.conn.schemadir, "*.ldif"))
        if ds_is_newer('1.3.6.0', instance=self._instance):
            file_list += glob.glob(os.path.join(self.conn.ds_paths.system_schema_dir, "*.ldif"))
        return file_list

    def file_to_ldap(self, filename):
        """Convert the given schema file name to its python-ldap format
        suitable for passing to ldap.schema.SubSchema()

        :param filename: the full path and filename of a schema file in ldif format
        :type filename: str
        """

        with open(filename, 'r') as f:
            ldif_parser = ldif.LDIFRecordList(f, max_entries=1)
        if not ldif_parser:
            return None
        ldif_parser.parse()
        if not ldif_parser.all_records:
            return None
        return ldif_parser.all_records[0][1]

    def file_to_subschema(self, filename):
        """Convert the given schema file name to its python-ldap format
        ldap.schema.SubSchema object

        :param filename: the full path and filename of a schema file in ldif format
        :type filename: str
        """

        ent = self.file_to_ldap(filename)
        if not ent:
            return None
        return ldap.schema.SubSchema(ent)

    def get_schema_csn(self):
        """Return the schema nsSchemaCSN attribute"""

        return self.get_attr_val_utf8('nsSchemaCSN')

    def add_attributetype(self, parameters):
        """Add an attribute type definition to the schema.

        :param parameters: an attribute type definition to add
        :type parameters: str
        """

        return self._add_schema_object(parameters, AttributeType)

    def add_objectclass(self, parameters):
        """Add an object class definition to the schema.

        :param parameters: an objectClass definition to add
        :type parameters: str
        """

        return self._add_schema_object(parameters, ObjectClass)

    def remove_attributetype(self, name):
        """Remove the attribute type definition from the schema.

        :param name: the name of the attributeType you want to remove.
        :type name: str
        """

        return self._remove_schema_object(name, AttributeType)

    def remove_objectclass(self, name):
        """Remove an objectClass definition from the schema.

        :param name: the name of the objectClass you want to remove.
        :type name: str
        """
        return self._remove_schema_object(name, ObjectClass)

    def edit_attributetype(self, name, parameters):
        """Edit the attribute type definition in the schema

        :param name: the name of the attribute type you want to edit.
        :type name: str
        :param parameters: an attribute type definition to edit
        :type parameters: str
        """

        return self._edit_schema_object(name, parameters, AttributeType)

    def edit_objectclass(self, name, parameters):
        """Edit an objectClass definition in the schema.

        :param name: the name of the objectClass you want to edit.
        :type name: str
        :param parameters: an objectClass definition to edit
        :type parameters: str
        """

        return self._edit_schema_object(name, parameters, ObjectClass)

    def get_objectclasses(self, include_sup=False, json=False):
        """Returns a list of ldap.schema.models.ObjectClass objects for all
        objectClasses supported by this instance.

        :param json: return the result in JSON format
        :type json: bool
        """

        return self._get_schema_objects(ObjectClass, include_sup=include_sup,
                                        json=json)

    def get_attributetypes(self, json=False):
        """Returns a list of ldap.schema.models.AttributeType objects for all
        attributeTypes supported by this instance.

        :param json: return the result in JSON format
        :type json: bool
        """

        return self._get_schema_objects(AttributeType, json=json)

    def get_matchingrules(self, json=False):
        """Return a list of the server defined matching rules

        :param json: return the result in JSON format
        :type json: bool
        """

        return self._get_schema_objects(MatchingRule, json=json)

    def query_matchingrule(self, mr_name, json=False):
        """Returns a single matching rule instance that matches the mr_name.
        Returns None if the matching rule doesn't exist.

        :param mr_name: the name of the matching rule you want to query.
        :type mr_name: str
        :param json: return the result in JSON format
        :type json: bool

        :returns: MatchingRule or None

        <ldap.schema.models.MatchingRule instance>
        """

        matching_rule = self._get_schema_object(mr_name, MatchingRule, json=json)

        if json:
            result = {'type': 'schema', 'mr': matching_rule}
            return result
        else:
            return matching_rule

    def query_objectclass(self, objectclassname, include_sup=False,
                          json=False):
        """Returns a single ObjectClass instance that matches objectclassname.
        Returns None if the objectClass doesn't exist.

        :param objectclassname: The name of the objectClass you want to query.
        :type objectclassname: str
        :param json: return the result in JSON format
        :type json: bool

        :returns: ObjectClass or None

        ex. query_objectclass('account')
        <ldap.schema.models.ObjectClass instance>
        """

        objectclass = self._get_schema_object(objectclassname, ObjectClass,
                                              include_sup=include_sup,
                                              json=json)

        if json:
            result = {'type': 'schema', 'oc': objectclass}
            return result
        else:
            return objectclass

    def query_attributetype(self, attributetypename, json=False):
        """Returns a tuple of the AttributeType, and what objectclasses may or
        must take this attributeType. Returns None if attributetype doesn't
        exist.

        :param attributetypename: The name of the attributeType you want to query
        :type attributetypename: str
        :param json: return the result in JSON format
        :type json: bool

        :returns: (AttributeType, Must, May) or None

        ex. query_attributetype('uid')
        ( <ldap.schema.models.AttributeType instance>,
         [<ldap.schema.models.ObjectClass instance>, ...],
         [<ldap.schema.models.ObjectClass instance>, ...] )
        """

        # First, get the attribute that matches name. We need to consider
        # alternate names. There is no way to search this, so we have to
        # filter our set of all attribute types.
        attributetype = self._get_schema_object(attributetypename, AttributeType, json=json)
        objectclasses = self.get_objectclasses()

        # Get the primary name of this attribute
        if json:
            attributetypenames = attributetype["names"]
        else:
            attributetypenames = attributetype.names

        # Build a set if they have may.
        may = []
        for attributetypename in attributetypenames:
            may.extend([oc for oc in objectclasses if attributetypename.lower() in
                        list(map(str.lower, oc.may))])
        # Build a set if they have must.
        must = []
        for attributetypename in attributetypenames:
            must.extend([oc for oc in objectclasses if attributetypename.lower() in
                         list(map(str.lower, oc.must))])

        if json:
            # convert Objectclass class to dict, then sort each list
            may = [vars(oc) for oc in may]
            must = [vars(oc) for oc in must]
            # Add normalized 'name' for sorting
            for oc in may:
                oc['name'] = oc['names'][0]
            for oc in must:
                oc['name'] = oc['names'][0]
            may = sorted(may, key=itemgetter('name'))
            must = sorted(must, key=itemgetter('name'))
            result = {'type': 'schema',
                      'at': attributetype,
                      'may': may,
                      'must': must}
            return result
        else:
            return str(attributetype), may, must

    def validate_syntax(self, basedn, _filter=None):
        """Create a validate syntax task

        :param basedn: Basedn to validate
        :type basedn: str
        :param _filter: a filter for entries to validate
        :type _filter: str

        :returns: an instance of Task(DSLdapObject)
        """

        task = SyntaxValidateTask(self._instance)
        task_properties = {'basedn': basedn}
        if _filter is not None:
            task_properties['filter'] = _filter
        task.create(properties=task_properties)

        return task


class SchemaLegacy(object):

    def __init__(self, conn):
        """@param conn - a DirSrv instance"""
        self.conn = conn
        self.log = conn.log

    def get_entry(self):
        """get the schema as an LDAP entry"""
        attrs = ['attributeTypes', 'objectClasses']
        return self.conn.search_s(DN_SCHEMA, ldap.SCOPE_BASE,
                                  'objectclass=*', attrs)[0]

    def get_subschema(self):
        """get the schema as a python-ldap SubSchema object"""
        return ldap.schema.SubSchema(self.get_entry().data)

    def list_files(self):
        """return a list of the schema files in the instance schemadir"""
        file_list = []
        file_list += glob.glob(self.conn.schemadir + "/*.ldif")
        if ds_is_newer('1.3.6.0', instance=self.conn):
            file_list += glob.glob(self.conn.ds_paths.system_schema_dir + "/*.ldif")
        return file_list

    def file_to_ldap(self, filename):
        """convert the given schema file name to its python-ldap format
        suitable for passing to ldap.schema.SubSchema()
        @param filename - the full path and filename of a schema file in ldif
        format"""

        with open(filename, 'r') as f:
            ldif_parser = ldif.LDIFRecordList(f, max_entries=1)
        if not ldif_parser:
            return None
        ldif_parser.parse()
        if not ldif_parser.all_records:
            return None
        return ldif_parser.all_records[0][1]

    def file_to_subschema(self, filename):
        """convert the given schema file name to its python-ldap format
        ldap.schema.SubSchema object
        @param filename - the full path and filename of a schema file in ldif
        format"""
        ent = self.file_to_ldap(filename)
        if not ent:
            return None
        return ldap.schema.SubSchema(ent)

    def add_schema(self, attr, val):
        """Add a schema element to the schema.
        @param attr - the attribute type to use e.g. attributeTypes or
                      objectClasses
        @param val the schema element definition to add"""
        self.conn.modify_s(DN_SCHEMA, [(ldap.MOD_ADD, attr, val)])

    def del_schema(self, attr, val):
        """Delete a schema element from the schema.
        @param attr - the attribute type to use e.g. attributeTypes or
                      objectClasses
        @param val - the schema element definition to delete"""
        self.conn.modify_s(DN_SCHEMA, [(ldap.MOD_DELETE, attr, val)])

    def add_attribute(self, *attributes):
        """Add an attribute type definition to the schema.
        @param attributes a single or list of attribute type defintions to add
        """
        return self.add_schema('attributeTypes', attributes)

    def add_objectclass(self, *objectclasses):
        """Add an object class definition to the schema.
        @param objectclasses a single or list of object class defintions to add
        """
        return self.add_schema('objectClasses', objectclasses)

    def get_schema_csn(self):
        """return the schema nsSchemaCSN attribute"""
        ents = self.conn.search_s(DN_SCHEMA, ldap.SCOPE_BASE,
                                  "objectclass=*", ['nsSchemaCSN'])
        ent = ents[0]
        return ent.getValue('nsSchemaCSN')

    def get_objectclasses(self, json=False):
        """Returns a list of ldap.schema.models.ObjectClass objects for all
        objectClasses supported by this instance.
        """
        attrs = ['objectClasses']
        results = self.conn.search_s(DN_SCHEMA, ldap.SCOPE_BASE,
                                     'objectclass=*', attrs)[0]
        if json:
            objectclasses = [vars(ObjectClass(oc)) for oc in
                results.getValues('objectClasses')]
            for oc in objectclasses:
                # Add normalized name for sorting
                oc['name'] = oc['names'][0]
            objectclasses = sorted(objectclasses, key=itemgetter('name'))
            result = {'type': 'list', 'items': objectclasses}
            return dump_json(result)
        else:
            objectclasses = [ObjectClass(oc) for oc in
                             results.getValues('objectClasses')]
            return objectclasses

    def get_attributetypes(self, json=False):
        """Returns a list of ldap.schema.models.AttributeType objects for all
        attributeTypes supported by this instance.
        """
        attrs = ['attributeTypes']
        results = self.conn.search_s(DN_SCHEMA, ldap.SCOPE_BASE,
                                     'objectclass=*', attrs)[0]

        if json:
            attributetypes = [vars(AttributeType(at)) for at in
                results.getValues('attributeTypes')]
            for attr in attributetypes:
                # Add normalized name for sorting
                attr['name'] = attr['names'][0]
            attributetypes = sorted(attributetypes, key=itemgetter('name'))
            result = {'type': 'list', 'items': attributetypes}
            return dump_json(result)
        else:
            attributetypes = [AttributeType(at) for at in
                results.getValues('attributeTypes')]
            return attributetypes

    def get_matchingrules(self, json=False):
        """Return a list of the server defined matching rules"""
        attrs = ['matchingrules']
        results = self.conn.search_s(DN_SCHEMA, ldap.SCOPE_BASE,
                                     'objectclass=*', attrs)[0]
        if json:
            matchingRules = [vars(MatchingRule(mr)) for mr in
                results.getValues('matchingRules')]
            for mr in matchingRules:
                # Add normalized name for sorting
                if mr['names']:
                    mr['name'] = mr['names'][0]
                else:
                    mr['name'] = ""
            matchingRules = sorted(matchingRules, key=itemgetter('name'))
            result = {'type': 'list', 'items': matchingRules}
            return dump_json(result)
        else:
            matchingRules = [MatchingRule(mr) for mr in
                         results.getValues('matchingRules')]
            return matchingRules

    def query_matchingrule(self, mr_name, json=False):
        """Returns a single matching rule instance that matches the mr_name.
        Returns None if the matching rule doesn't exist.

        @param mr_name - The name of the matching rule you want to query.

        return MatchingRule or None

        <ldap.schema.models.MatchingRule instance>
        """
        matchingRules = self.get_matchingrules()
        matchingRule = [mr for mr in matchingRules if mr_name.lower() in
                        list(map(str.lower, mr.names))]
        if len(matchingRule) != 1:
            # This is an error.
            if json:
                raise ValueError('Could not find matchingrule: ' + mr_name)
            else:
                return None
        matchingRule = matchingRule[0]
        if json:
            result = {'type': 'schema', 'mr': vars(matchingRule)}
            return dump_json(result)
        else:
            return matchingRule

    def query_objectclass(self, objectclassname, json=False):
        """Returns a single ObjectClass instance that matches objectclassname.
        Returns None if the objectClass doesn't exist.

        @param objectclassname - The name of the objectClass you want to query.

        return ObjectClass or None

        ex. query_objectclass('account')
        <ldap.schema.models.ObjectClass instance>
        """
        objectclasses = self.get_objectclasses()

        objectclass = [oc for oc in objectclasses if objectclassname.lower() in
                       list(map(str.lower, oc.names))]
        if len(objectclass) != 1:
            # This is an error.
            if json:
                raise ValueError('Could not find objectcass: ' + objectclassname)
            else:
                return None
        objectclass = objectclass[0]
        if json:
            result = {'type': 'schema', 'oc': vars(objectclass)}
            return dump_json(result)
        else:
            return objectclass

    def query_attributetype(self, attributetypename, json=False):
        """Returns a tuple of the AttributeType, and what objectclasses may or
        must take this attributeType. Returns None if attributetype doesn't
        exist.

        @param attributetypename - The name of the attributeType you want to
        query

        return (AttributeType, Must, May) or None

        ex. query_attributetype('uid')
        ( <ldap.schema.models.AttributeType instance>,
         [<ldap.schema.models.ObjectClass instance>, ...],
         [<ldap.schema.models.ObjectClass instance>, ...] )
        """
        # First, get the attribute that matches name. We need to consider
        # alternate names. There is no way to search this, so we have to
        # filter our set of all attribute types.
        objectclasses = self.get_objectclasses()
        attributetypes = self.get_attributetypes()

        attributetype = [at for at in attributetypes
                         if attributetypename.lower() in
                         list(map(str.lower, at.names))]
        if len(attributetype) != 1:
            # This is an error.
            if json:
                raise ValueError('Could not find attribute: ' + attributetypename)
            else:
                return None
        attributetype = attributetype[0]
        # Get the primary name of this attribute
        attributetypename = attributetype.names[0]
        # Build a set if they have may.
        may = [oc for oc in objectclasses if attributetypename.lower() in
               list(map(str.lower, oc.may))]
        # Build a set if they have must.
        must = [oc for oc in objectclasses if attributetypename.lower() in
                list(map(str.lower, oc.must))]

        if json:
            # convert Objectclass class to dict, then sort each list
            may = [vars(oc) for oc in may]
            must = [vars(oc) for oc in must]
            # Add normalized 'name' for sorting
            for oc in may:
                oc['name'] = oc['names'][0]
            for oc in must:
                oc['name'] = oc['names'][0]
            may = sorted(may, key=itemgetter('name'))
            must = sorted(must, key=itemgetter('name'))
            result = {'type': 'schema',
                      'at': vars(attributetype),
                      'may': may,
                      'must': must}
            return dump_json(result)
        else:
            return (attributetype, must, may)

class Resolver(object):
    def __init__(self, schema_attrs):
        self.attr_map = {}
        for attr in schema_attrs:
            for name in attr.names:
                self.attr_map[name.lower()] = attr
        # done

    def resolve(self, attr_in):
        attr_in_l = attr_in.lower()
        if attr_in_l in self.attr_map:
            return self.attr_map[attr_in_l].names[0]
        else:
            return attr_in_l



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest
import ldap

from lib389._constants import DEFAULT_SUFFIX, INSTALL_LATEST_CONFIG

from lib389.cli_conf.backend import backend_create
from lib389.cli_idm.initialise import initialise
from lib389.cli_idm.user import create, modify

from lib389.cli_base import LogCapture, FakeArgs
from lib389.tests.cli import topology

from lib389.utils import ds_is_older
pytestmark = pytest.mark.skipif(ds_is_older('1.4.0'), reason="Not implemented")

# Topology is pulled from __init__.py
def test_user_modify(topology):
    be_args = FakeArgs()

    be_args.be_name = 'userRoot'
    be_args.suffix = DEFAULT_SUFFIX
    be_args.parent_suffix = None
    be_args.create_entries = False
    backend_create(topology.standalone, None, topology.logcap.log, be_args)

    # And add the skeleton objects.
    init_args = FakeArgs()
    init_args.version = INSTALL_LATEST_CONFIG
    initialise(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, init_args)

    # Check that our modify parser works. Modify statements are such as:
    # "add:attr:value". Replace is the exception as "replace:attr:old:new"

    # Check bad syntax
    modify_args = FakeArgs()
    modify_args.selector = "demo_user"
    modify_args.changes = ["tnaohtnsuahtnsouhtns"]

    with pytest.raises(ValueError):
        modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["add:attr:"]
    with pytest.raises(ValueError):
        modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["add:attr"]
    with pytest.raises(ValueError):
        modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["replace::"]
    with pytest.raises(ValueError):
        modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["replace:attr::new"]
    with pytest.raises(ValueError):
        modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["delete:attr:old:new"]
    with pytest.raises(ValueError):
        modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    # Check that even a single bad value causes error
    modify_args.changes = ["add:description:goodvalue", "add:attr:"]
    with pytest.raises(ValueError):
        modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    # check good syntax
    modify_args.changes = ["add:description:testvalue"]
    modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["replace:description:newvalue"]
    modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["delete:description:newvalue"]
    modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["add:description:testvalue"]
    modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    modify_args.changes = ["delete:description:"]
    modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)

    # check mixed type, with multiple actions

    modify_args.changes = ["add:objectclass:nsMemberOf", "add:description:anothervalue"]
    modify(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, modify_args)




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from argparse import ArgumentTypeError
from lib389.plugins import ContentSyncPlugin
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add_attr
from lib389.cli_base import CustomHelpFormatter

arg_to_attr = {
    'allow_openldap': 'syncrepl-allow-openldap',
    'queue_max_size': 'syncrepl-queue-max-size',
    'max_concurrent': 'syncrepl-max-concurrent',
}

def check_queue_size(value):
    ivalue = int(value)
    if 100 <= ivalue <= 100000:
        return str(ivalue)
    else:
        raise ArgumentTypeError(f"{value} should be in the range [100, 100000]")

def contentsync_edit(inst, basedn, log, args):
    log = log.getChild('contentsync_edit')
    plugin = ContentSyncPlugin(inst)
    generic_object_edit(plugin, log, args, arg_to_attr)


def contentsync_add(inst, basedn, log, args):
    log = log.getChild('contentsync_add')
    plugin = ContentSyncPlugin(inst)
    generic_object_add_attr(plugin, log, args, arg_to_attr)


def _add_parser_args(parser):
    parser.add_argument('--allow-openldap', choices=['on', 'off'], type=str.lower,
                        help='Allows openldap servers to act as read only consumers of this server via syncrepl')
    parser.add_argument('--queue-max-size', type=check_queue_size,
                        help='Limits the number of entries not yet processed (range [100, 100000])')
    parser.add_argument('--max-concurrent', type=str,
                        help='Limits the number of persistent searches running at the same time')

def create_parser(subparsers):
    contentsync_parser = subparsers.add_parser('contentsync', help='Manage and configure Content Sync Plugin (aka syncrepl)', formatter_class=CustomHelpFormatter)
    subcommands = contentsync_parser.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, ContentSyncPlugin)

    edit = subcommands.add_parser('set', help='Edit the plugin settings', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=contentsync_edit)
    _add_parser_args(edit)

    addp = subcommands.add_parser('add', help='Add attributes to the plugin', formatter_class=CustomHelpFormatter)
    addp.set_defaults(func=contentsync_add)
    _add_parser_args(addp)
